// user.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"



#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



extern CFilerDlg * gMaindialog;
extern char * gMyDirectory;
extern CSciCtrl * gScintilla;
extern bool gTesting;
extern bool gPlaying;
extern bool gStepping;
extern bool gRecording;
extern macro * gCurmac;
extern macro * gMacros;
extern CAtom gAtom;
extern bool gLaunchedfirst;
extern HWND gDest;
extern char * gPathtoshow;
//////////////////////////////////////////////////////////////////////////////
// Calling another program

void CFilerDlg::WindowsOpen(const char * fpath)
{
	ShellExecute(NULL,"open",fpath,NULL,NULL,SW_SHOWDEFAULT);
}

void CFilerDlg::ExplorerOpen(const char * fpath)
{
	ShellExecute(NULL,"explore",fpath,NULL,NULL,SW_SHOWDEFAULT);
}


void CFilerDlg::SciTEOpen(const char * path,bool activate)
{
	if (gDest==0 && gLaunchedfirst){
		char SciTEPath[MAX_PATH];
		char SciTEparams[MAX_PATH];
		char parameters[MAX_PATH+50];
		char cmdlineformat[]="-director.hwnd=%u %s \"%s\"";
		BuildPath("scitepath",SciTEPath,"scite.exe");
		GetIniWorkString("sciteparams",SciTEparams);
		sprintf(parameters,cmdlineformat,(UINT)m_hWnd,SciTEparams,path);
		ShellExecute(NULL,NULL,SciTEPath,parameters,gMyDirectory,SW_SHOWDEFAULT);
		return;
	} else {
		ToSendMessage(SCD_OPEN,path);
		if (activate)
			BringSciTEToForeground();
	}
}


void CFilerDlg::BringSciTEToForeground() 
{
	::SetForegroundWindow((HWND)::GetWindowLong(gDest,GWL_USERDATA));
}

/////////////////////////////////////////////////////////////////////////////
// CFilerDlg Talking with User



struct receive_comm {
	int command_code;
	char * command_text;
	int subcommand;
};

static receive_comm to_receive[]={
		{SCD_IDENTIFY,"identity",0},
		{SCD_CLOSING,"closing",0},
		{SCD_OPENED,"opened",0},
		{SCD_FILENAME,"filename",0},
		{SCD_SAVED,"saved",0},
		{SCD_SWITCHED,"switched",0},
		{SCD_PROPERTY,"property",0},
		{SCD_MACRO,"macro",1},
			{SCD_MACRO_RUN,"run",0},
			{SCD_MACRO_GETLIST,"getlist",0},
			{SCD_MACRO_RECORD,"record",0},
			{SCD_MACRO_STOPRECORD,"stoprecord",0},
			{SCD_MACRO_INTINFO,"intinfo",0},
			{SCD_MACRO_STRINGINFO,"stringinfo",0},
		{0,""},
};


static char * TypeFromCommand(char * command, int * type){
	char *cmd =command;
	char * cmdend =strchr(cmd,':');
	if (cmdend!=NULL)
		*cmdend='\0';
	else
		cmdend=cmd+lstrlen(cmd);
	
	int i=0;
	do{
		if (!lstrcmp(to_receive[i].command_text,cmd)){
			if (to_receive[i].subcommand==1){
				cmd=cmdend+1;
				cmdend =strchr(cmd,':');
				if (cmdend!=NULL)
					*cmdend='\0';
			}else
				break;
		}
		i++;
	}while (to_receive[i].command_code!=0);
	*type=to_receive[i].command_code;
	return cmdend+1;
}




//Receiving data
void CFilerDlg::DoLog(const char * log1,const char * log2)
{
	if (logging) {
		logfile.WriteString(log1);		
		logfile.WriteString(log2);
		logfile.WriteString("\n");
	}
}

BOOL CFilerDlg::OnCopyData(CWnd* pWnd, COPYDATASTRUCT* pcds) 
{
	if (pcds->dwData==0){

		char * data=(char *)pcds->lpData;
		unsigned int nCopy=pcds->cbData;
		char * datacopy=new char[nCopy+1];
		strncpy(datacopy,data, nCopy);
		datacopy[nCopy] = '\0';

		DoLog("<<  ",datacopy);

		char * pCommand=datacopy;

        if (*datacopy == ':') {
            // There is a return address
            char *colon = strchr(datacopy+1, ':');
            if (colon) {
                *colon = '\0';
                wCorrespondent = reinterpret_cast<HWND>(atoi(datacopy+1));
                pCommand = colon+1;
            }
        }
                
		int type;
		char * arg=TypeFromCommand(pCommand,&type);
		UnSlash(arg);

		SetStatusIcon(0,type!=SCD_CLOSING);


		switch(type){
			case 0 :
				break;
			case SCD_IDENTIFY:	//SciTE is identifying :
				gDest=reinterpret_cast<HWND>(atoi(arg));
				ToSendMessage(SCD_MACROENABLE,"1");
				gAtom.Add((unsigned int)gDest);
				break;
			case SCD_CLOSING:	//SciTE is closing : I close, too if I was launched by SciTE
				gDest=0;
				if (!gLaunchedfirst){
					CleanBeforeClosing();
					CDialog::OnCancel();
				} else {
					gAtom.Delete();
					SetTestMode(true);
					UpdateViews(-1,-1);
				}
				break;
			case SCD_OPENED:	//SciTE opened a new file :ShowPath
			case SCD_SWITCHED:	//SciTE switched buffers : ShowPath
				lstrcpy(gPathtoshow,arg);
				TryToShowPath();
				break;
			case SCD_SAVED:		//SciTE saved a file : Refresh
				m_tree.Refresh(arg);
				break;
			case SCD_PROPERTY:	//SciTE sends me a property
				m_proptree.AddProperty(arg);
				break;
			case SCD_MACRO_RUN: {
				if (gTesting)
					ToggleTestMode();
				char * argbeg=strchr(arg,'(');
				if (argbeg!=NULL) 
					RunMacroWithArgs(arg,NULL);
				 else
					PlayMacro(arg);
				break;
				}
			case SCD_MACRO_GETLIST:
					BuildMacroList();
				break;
			case SCD_MACRO_RECORD:
				RecordCommand(arg);
				break;
			case SCD_MACRO_STOPRECORD:
				if (gRecording)
				StopRecordMacro();
				break;
			case SCD_MACRO_INTINFO:
				sci_int_info=atoi(arg);
				gotinfo=true;
				break;
			case SCD_FILENAME:
			case SCD_MACRO_STRINGINFO:
				sci_text_info=arg;
				gotinfo=true;
				break;
		}
		wCorrespondent=0;
		delete [] datacopy;
		return true;
	}
	return CDialog::OnCopyData(pWnd, pcds);
}




//Sending data


struct send_comm {
	int command_code;
	char * command_text;
};


static send_comm to_send[]={
		{SCD_IDENTIFY,"identity:"},
		{SCD_OPEN,"open:"},
		{SCD_LOADSCRIPT,"extender:dofile "},
		{SCD_EXTENDER,"extender:"},
		{SCD_CLOSE,"close:"},
		{SCD_QUIT,"quit:"},
		{SCD_SAVEAS,"saveas:"},
		{SCD_EXPORTASHTML,"exportashtml:"},
		{SCD_EXPORTASPDF,"exportaspdf:"},
		{SCD_EXPORTASRTF,"exportasrtf:"},
		{SCD_MENUCOMMAND,"menucommand:"},
		{SCD_CLOSING,"closing:"},
		{SCD_ENUMPROPERTIES,"enumproperties:"},
		{SCD_SETPROPERTY,"property:"},
		{SCD_MACROLIST,"macrolist:"},
		{SCD_CURRENTMACRO,"currentmacro:"},
		{SCD_ASKFILENAME,"askfilename:"},
		{SCD_ASKPROPERTY,"askproperty:"},
		{SCD_REPLACEALL,"replaceall:"},
		{SCD_FIND,"find:"},
		{SCD_MACROENABLE,"macroenable:"},
		{SCD_MACROCOMMAND,"macrocommand:"},
		{SCD_OUTPUT,"output:"},
		{SCD_OUTPUTSIZE,"outputsize:"},

		{SCD_CWD,"cwd:"},
		{SCD_EXPORTASLATEX,"exportaslatex:"},
		{SCD_EXPORTASXML,"exportasxml:"},
		{SCD_GOTO,"goto:"},
		{SCD_INSERT,"insert:"},
		{SCD_LOADSESSION,"loadsession:"},
		{0,""},
};

static char * CommandFromType(int type){
	int i=0;
	do{
		if (to_send[i].command_code==type)
			break;
		i++;
	}while (to_send[i].command_code!=0);
	return (to_send[i].command_text);
}

int ScanSubmenu(HMENU hmenu,const char * text)
{
	MENUITEMINFO minfo;
	char menutext[256];
	int tlen=lstrlen(text);
	int hmenucount=::GetMenuItemCount(hmenu);
	for (int i=0;i<=hmenucount-1;i++) {
		::ZeroMemory(&minfo,sizeof(minfo));
		minfo.cbSize = sizeof(MENUITEMINFO); 
		minfo.fMask=MIIM_TYPE | MIIM_ID | MIIM_SUBMENU;
		minfo.fType = MFT_STRING; 
		minfo.dwTypeData=menutext;
		minfo.cch=256;
		*menutext='\0';
		if (::GetMenuItemInfo(hmenu,i,true,&minfo)) {
			if (minfo.hSubMenu!=0) {
				int res=ScanSubmenu(minfo.hSubMenu,text);
				if (res!=0) return res;
			} else {
				char * p=menutext;
				char * q=p;
				while (*p) {
					if (*p=='\t')
						break;
					else if (*p=='&')
						p++;
					else
						*q++=*p++;
				}
				*q='\0';
				if (!strncmp(menutext,text,tlen))
					return minfo.wID;
			}
		}
	}
	return 0;
}

int FindSciTEmenu(const char * text)
{
	HMENU hmainmenu=::GetMenu((HWND)::GetWindowLong(gDest,GWL_USERDATA));
	int menucount=::GetMenuItemCount(hmainmenu);
	for (int i=0;i<=menucount-1;i++) {
		int id=ScanSubmenu(::GetSubMenu(hmainmenu,i),text);
		if (id!=0) return id;
	}
	return 0;
}

int FindWmenu(const char * text,int arg2)
{
	HMENU hmainmenu=::GetMenu((HWND)arg2);
	int menucount=::GetMenuItemCount(hmainmenu);
	for (int i=0;i<=menucount-1;i++) {
		int id=ScanSubmenu(::GetSubMenu(hmainmenu,i),text);
		if (id!=0) return id;
	}
	return 0;
}

void CFilerDlg::ToSendMessage(int typ,const char * arg1 /*=NULL*/,const char * arg2/*=NULL*/)
{	
	const char * arg=arg1;
	char numbuff[20];
	if (typ>=FLX_MESSAGES) {
		FilerxMessage(typ,arg1,arg2);
		return;
	}

	if (typ==SCD_OPEN && gDest==0 && gLaunchedfirst) {
		SciTEOpen(arg1);
		DoLog(" >> open:",arg1);
		return;
	} else if (typ==SCD_MENUCOMMAND) {
		if (gDest==0) return;
		int foundmenu=FindSciTEmenu(arg1);
		if (foundmenu==0) return;
		itoa(foundmenu,numbuff,10);
		arg=numbuff;
	} else if (typ==SCD_OUTPUTSIZE || typ==SCD_GOTO) {
		itoa((int) arg,numbuff,10);
		arg=numbuff;
	} else if (typ==SCD_MENUCOMMANDW) {
		int foundmenu=FindWmenu(arg1,(int)arg2);
		if (foundmenu!=0)
			::SendMessage((HWND) arg2 ,WM_COMMAND,(UINT)foundmenu,0);
		return;
	} else if (typ==SCD_MINIMIZE || typ==SCD_RESTORE) {
		if (gDest!=0) {
			int howtoshow=(typ==SCD_MINIMIZE ? SW_MINIMIZE : SW_RESTORE);
			HWND scitewin=(HWND)::GetWindowLong(gDest,GWL_USERDATA);
			::ShowWindow(scitewin,howtoshow);
		}
	}
	HWND wDestination=gDest;
	char cCorres[10];
	*cCorres='\0';
    if (wCorrespondent != 0) {
        HWND wDestination = wCorrespondent;
        sprintf(cCorres,":%i:",wDestination);
	}
    
	//If I have somebody to talk to :

	if (gDest!=0 || wCorrespondent != 0){
		char * cmdtxt=CommandFromType(typ);

		//Build szMessage
		Cstr szMessage(cCorres);
		szMessage+=cmdtxt;
		if (arg!=NULL){
			szMessage.SlashAppend(arg);
			if (arg2!=NULL) {
				szMessage+="\\0";
                szMessage.SlashAppend(arg2);
			}
		}
		
		DoLog(" >> ",szMessage.buf);

		//Send message
		COPYDATASTRUCT cds;
        cds.dwData = 0;
		cds.lpData = szMessage.buf;
		cds.cbData = szMessage.len+1;
		::SendMessage(wDestination ,WM_COPYDATA,(UINT)m_hWnd,(long)&cds);
	}
}


void CFilerDlg::ToSendMacroCommand(int cmd,int p1,char * t,char * params)
{
	char itoabuf1[14];
	char * arg1="\0";
	char itoabuf2[14];
	char * arg2="\0";
	char prms[4];

	strncpy(prms,params,3);
	*(prms+3)='\0';
	
	if (gDest==NULL)
		return;
	
	if (*(params+1)=='I'){
		itoa(p1,itoabuf1,10);
		arg1=itoabuf1;
	}else if (p1!=0)
		arg1=(char *) p1;

	if (*(params+2)=='I'){
		itoa((int)t,itoabuf2,10);
		arg2=itoabuf2;
	}else if (t!=0)
		arg2=t;

	char * arg=new char [14+4+3+lstrlen(arg1)+lstrlen(arg2)];
	sprintf(arg, "%i;%s;%s;%s",cmd,prms,arg1,arg2);
	ToSendMessage(SCD_MACROCOMMAND,arg);
	delete []arg;
}



//Called by expression evaluator to retrieve SciTE/Scintilla state

int CFilerDlg::GetInfo(int code,char *params,int arg,int arg2)
{

	if (code>=FLX_MESSAGES) {
		return GetFilerxInfo(code,params,arg,arg2);
	}

	if (gDest!=NULL){
		gotinfo=false;
		if (code==SCD_ASKFILENAME)
			ToSendMessage(code);
		else if (code==SCD_ASKPROPERTY || code==SCD_MACROCOMMAND)
			ToSendMessage(code,(const char *)arg);
		else
			ToSendMacroCommand(code,arg,(char *)arg2,params);
	
		while (gotinfo==false); // waits an answer, maybe forever !!!
	}else{
		sci_text_info="";
		sci_int_info=0;
	}

	if (*params=='I')
			return sci_int_info;
	else
		return (int)sci_text_info.buf;

	return 0;
}

void CFilerDlg::SendText(const char * txt)
{
	int cmd=m_scintilla.SCI_ADDTEXT;
	ToSendMacroCommand(cmd,lstrlen(txt),(char *)txt,"0IS");
}

void CFilerDlg::SendWithToolTip(const char * txt)
{
	char buff[100];
	char * p=(char *)txt;
	char * pcopy=buff;
	while ( *p && *p!='(') 
		*pcopy++=*p++;
	*pcopy++='(';*pcopy++='\0';
	SendText((const char *) buff);
	int cpos=GetInfo(m_scintilla.SCI_GETCURRENTPOS,"I00",0,0);
	ToSendMacroCommand(m_scintilla.SCI_CALLTIPSHOW,cpos,(char *)txt,"0IS");
	BringSciTEToForeground();
}


/////////////////////////////////////////////////////////////////////////////
//  PLAYING MACROS

UINT PlayThread( LPVOID pParam )
{
	gMaindialog->SetStatusIcon(1,true);
	Play((macro *)pParam);
	gPlaying=false;

	if (gTesting){
		gScintilla->AfterPlay();
	}
	gMaindialog->SetStatusIcon(1,false);
	gMaindialog->DoSplash("");
    return 0;   // thread completed successfully
}

UINT DebugThread( LPVOID pParam )
{
	gMaindialog->SetStatusIcon(1,true);
	gStepping=true;
	Debug((macro *)pParam);
	gPlaying=false;
	gStepping=false;
	gScintilla->UnlockWindowUpdate();
	if (gTesting){
		gScintilla->AfterPlay();
	}
	gMaindialog->SetStatusIcon(1,false);
	gMaindialog->DoSplash("");
    return 0;   // thread completed successfully
}

void CFilerDlg::DebugOrPlay(bool todebug,char * m)
{
	if (!gPlaying) { 
		if (m!=NULL)
			gCurmac=m_mactree.FindMacroByName(m);
		if (gCurmac!=NULL) {
			if ((gCurmac->flags & 1) != 1) {
				if (gTesting)
					m_scintilla.BeforePlay();
				else
					ToSendMessage(SCD_CURRENTMACRO,gCurmac->name);
				gPlaying=true;
				if (todebug)
						AfxBeginThread(DebugThread, gCurmac);
				else
#ifdef _DEBUG
						AfxBeginThread(PlayThread, gCurmac);
#else
						AfxBeginThread(PlayThread, gCurmac,10);
#endif
			}
		}
	}
}

//Play /test a macro :
void CFilerDlg::PlayMacro(char * m /*=NULL*/){
	DebugOrPlay(false,m);
}

//Debug a macro :
void CFilerDlg::DebugMacro(char * m /*=NULL*/){
	DebugOrPlay(true,m);
}


/////////////////////////////////////////////////////////////////////////////
//  RECORDING MACROS


static unsigned int readnum(char ** t)
{
    char * argend=strchr(*t,';');    // find ';' 
    *argend='\0';                    // replace by null
    unsigned int v=atoi(*t);                // read value
    *t=argend+1;                    // update pointer
    return(v);                        // return value
}


void CFilerDlg::RecordCommand(char * c)
{	
    char * nextarg=(char *)c;
    unsigned int cmd;
    unsigned int wParam;
	unsigned int lParam=0;

	if (*c=='\0'){
		if (gRecording)
		StopRecordMacro();
		return;
	}
	if (!gRecording)
		StartRecordMacro();

	cmd=readnum(&nextarg);
    wParam=readnum(&nextarg);

	if (readnum(&nextarg)!=0)
		//format is  : "<message>;<wParam>;1;<text>"
		lParam=(unsigned int)nextarg;

	//else format is  : "<message>;<wParam>;0;"
	
	Record(gCurmac,cmd,wParam,lParam);
}

void CFilerDlg::StartRecordMacro(bool forcetesting)
{
	if (gPlaying)
		return;
	if (forcetesting && !gTesting)
		ToggleTestMode(false); 
	gCurmac=CreateMacro();
	gRecording=true;
	SetStatusIcon(2,true);
	if (gTesting) {
		m_scintilla.StartStopRecord(1);
	}
	int vmode2=((viewmode2==VIEW2_FILES)?VIEW2_OUTPUT:-1);
	UpdateViews(VIEW_MACROS,vmode2);

}

void CFilerDlg::StopRecordMacro()
{
	Record(gCurmac,0,0,0);
	gRecording=false;
	SetStatusIcon(2,false);
	if (gTesting) 
		m_scintilla.StartStopRecord(0);

	m_mactree.AddRecordedMacro();
	ToSendMessage(SCD_CURRENTMACRO,gCurmac->name);
	UpdateViews(VIEW_MACROS,-1);
}


///////////////////////////////////////////////////////////////////////
// Event Macros

void CFilerDlg::RunMenuMacro(const char * arg)
{
	char * torun=new char [chosenmacro.len+strlen(arg)+5];
	sprintf(torun,"%s(\"%s\")",chosenmacro.buf,arg);
	bool b=RunMacroWithArgs(torun,NULL);
	delete [] torun;
	return;
}


bool CFilerDlg::RunMacroWithArgs(const char * macname,const char * macargs)
{
	bool success=false;
// create calling macro
	Cstr macrotext("Macro void ___TTT___\ncall ");
	macrotext.SlashAppend(macname); // arguments included, if macargs==NULL
	if (macargs!=NULL) {
		macrotext+="(";
		macrotext.SlashAppend(macargs);		
		macrotext+=")";
	}
	macrotext+="\n\n";

// compile calling macro
	macro * tempmacro=CreateMacro();
	char * mend;
	Compile(tempmacro,gMacros,macrotext.buf,&mend);
// execute calling macro
	if ((tempmacro->flags & 255)==good_macro) {
		bool oldplaying=gPlaying;
		//if (gTesting) m_scintilla.BeforePlay();
		gPlaying=true;
		Play(tempmacro);
		gPlaying=oldplaying;
		success=true;
	}
// delete calling macro
	DeleteMacro(tempmacro);

	return success;
}




/////////////////////////////////////////////////////////////////////////////
//  LIST MACROS

void CFilerDlg::BuildMacroList() {

	Cstr buffer(500);

	macro * m=gMacros;
	while (m!=NULL){
		if ((m->flags & 1) !=1 && (m->flags & erased_macro)==0){ //not bad or needing parms
			buffer+=m->name;
			buffer+=" ";
		}
		m=m->next;
	}
	if (buffer.len!=0)
		gMaindialog->ToSendMessage(SCD_MACROLIST,buffer.buf);
}

/////////////////////////////////////////////////////////////////////////////
//  TEST MODE

void CFilerDlg::SetTestMode(bool test)
{
	gTesting=test;
	SetCommunicator(gTesting?(CMacroCommunicator *) gScintilla:(CMacroCommunicator *) gMaindialog);
	PrepareCopyData(!gTesting);
}
void CFilerDlg::EnterTestMode()
{
	SetTestMode(true);
	UpdateViews(VIEW_MACROS,VIEW2_OUTPUT);
}

void CFilerDlg::ExitTestMode()
{
	SetTestMode(false);
	UpdateViews(VIEW_MACROS,VIEW2_SOURCE);
}


void CFilerDlg::ToggleTestMode(bool forceupdate/*=true*/)
{ 
	gTesting=!gTesting;
	if (!gTesting && gDest==0) {
		SciTEOpen("",false);
		SetForegroundWindow();
	}
	SetTestMode(gTesting);
	if (forceupdate)
		UpdateViews(-1,-1);

}



