// FileListCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFilerDlg * gMaindialog;
extern CImgList * gImgList;

/////////////////////////////////////////////////////////////////////////////
// CFileListCtrl


CFileListCtrl::CFileListCtrl()
{
	curfilternum=0;
	curfilterlist="";
	m_pDragImage = NULL;
	m_bDragging=false;
}

CFileListCtrl::~CFileListCtrl()
{
	if(m_pDragImage)
		delete m_pDragImage;
}

void CFileListCtrl::DeleteAllItems()
{
  CListCtrl::DeleteAllItems();
}

bool CFileListCtrl::InFilter(const char * s)
{
	if (curfilternum<1) return true;
	char * s_suffix=strrchr(s,'.');
	if (s_suffix!=NULL){
		int suflen=lstrlen(s)-(s_suffix-s);
		char * filterlist=curfilterlist.buf;
		while (1) {
			char * foundsuffix=strstr(filterlist,s_suffix);
			if (foundsuffix==NULL) return false;
			if (*(foundsuffix+suflen)==' ') return true;
			filterlist=foundsuffix+1;
		}
	}
	return false;
}

void CFileListCtrl::DisplayList(TreeNode * pNode)
{
	DeleteAllItems();
	int it;
	TreeNode* pChild=pNode->m_Files;
    while (pChild!=NULL) {
		char * filename=pChild->m_strName;
		if (InFilter(filename)){
			it=InsertItem(0,filename,pChild->m_icon);
			DWORD fattr=GetFileAttributes(pChild->m_strPath);
			if (fattr & FILE_ATTRIBUTE_READONLY)
				SetItemState(it,LVIS_CUT,LVIS_CUT);
			SetItemData(it,(unsigned long)pChild);
		}
		pChild=pChild->m_next;
	}		
	if (gMaindialog->m_toolbar!=NULL)
		gMaindialog->m_toolbar->DisplayIconList(pNode->m_Files);
}


BEGIN_MESSAGE_MAP(CFileListCtrl, CListCtrl)
	//{{AFX_MSG_MAP(CFileListCtrl)
	ON_WM_LBUTTONDBLCLK()
	ON_NOTIFY_REFLECT(LVN_BEGINDRAG, OnBegindrag)
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFileListCtrl message handlers

void CFileListCtrl::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	char selpath[MAX_PATH];
	GetFirstSelectedPath(selpath);
	if (IsProject(selpath)){
		gMaindialog->m_ptree.AddProject(selpath);//Load Project file
		gMaindialog->UpdateViews(VIEW_PROJS,-1);
	}else if (IsDSP(selpath)){
		gMaindialog->m_ptree.ImportDSP(selpath);//Load DSP file
		gMaindialog->UpdateViews(VIEW_PROJS,-1);
	}else if (IsMacros(selpath)){
		gMaindialog->m_mactree.LoadMacroFile(selpath);//Load Macro file
		gMaindialog->UpdateViews(VIEW_MACROS,-1);
	} else {
		gMaindialog->SciTEOpen(selpath,true); 
	}
	CListCtrl::OnLButtonDblClk(nFlags, point);
}

void CFileListCtrl::SelectFile(char * name)
{
	LVFINDINFO lvf;
	lvf.flags=LVFI_STRING;
	lvf.psz=name;

	int founditem=FindItem(&lvf);
	if (founditem>=0){
		SetItemState(founditem,LVIS_SELECTED ,LVIS_SELECTED );
		EnsureVisible(founditem,0);
	}
	if (gMaindialog->m_toolbar!=NULL)
		gMaindialog->m_toolbar->Synchronize(&lvf);
}

void CFileListCtrl::GetFirstSelectedPath(char * p) 
{
	SelectedPos = GetFirstSelectedItemPosition();
	GetNextSelectedPath(p);
}

void CFileListCtrl::GetNextSelectedPath(char * p) 
{
	*p='\0';
	int SelectedItem=-1;
	if (SelectedPos != NULL) 
		SelectedItem = GetNextSelectedItem(SelectedPos);

	if (SelectedItem!=-1){
		TreeNode* pNode = (TreeNode*)GetItemData(SelectedItem);
		lstrcpy(p,pNode->m_strPath);
	}
}

	
//////////////////////////////////////////////////////////////////////
// Drag And Drop

void CFileListCtrl::OnBegindrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;
	m_nDragIndex = ((NM_LISTVIEW *)pNMHDR)->iItem;
	POINT pt;
	pt.x = 8;
	pt.y = 8;
	if(m_pDragImage) {
		delete m_pDragImage;
		m_pDragImage=NULL;
	}
	
	m_pDragImage = CreateDragImage (m_nDragIndex, &pt);
	m_pDragImage->BeginDrag (0, CPoint (8, 8));
	m_pDragImage->DragEnter (GetDesktopWindow (), ((NM_LISTVIEW *)pNMHDR)->ptAction);
	m_bDragging = TRUE;
	gMaindialog->m_ptree.m_hDropItem = NULL;
	SetCapture ();
	
	*pResult = 0;
}

void CFileListCtrl::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (m_bDragging)
	{
		ReleaseCapture ();
		m_bDragging = FALSE;
		m_pDragImage->DragLeave (GetDesktopWindow ());
		m_pDragImage->EndDrag ();
		CPoint pt (point);
		ClientToScreen (&pt);
		CWnd * w=WindowFromPoint (pt);
		if (w==(CWnd *)(&gMaindialog->m_ptree))
			gMaindialog->m_ptree.DropItemFromList ();
		else if (w->GetParent()==(CWnd *)(gMaindialog->m_toolbar))
			gMaindialog->m_toolbar->DropItemFromList (w);

	}	
	CListCtrl::OnLButtonUp(nFlags, point);
}

void CFileListCtrl::OnMouseMove(UINT nFlags, CPoint point) 
{
	CPoint pt (point);
	if (m_bDragging)
	{
		ClientToScreen (&pt);
		m_pDragImage->DragMove (pt);
		m_pDragImage->DragShowNolock (FALSE);
		gMaindialog->m_ptree.SetDropMark(pt);
		char selpath[MAX_PATH];
		GetFirstSelectedPath(selpath);
		if (gMaindialog->m_ptree.m_hDropItem==NULL && IsIcon(selpath))
			if (gMaindialog->m_toolbar!=NULL)
				gMaindialog->m_toolbar->SetDropMark(pt);
		m_pDragImage->DragShowNolock (TRUE);
	} else {
		int indx=HitTest(point,NULL);
		if (indx!=-1) {
			TreeNode* pNode = (TreeNode*)GetItemData(indx);

			CFileFind ffind;
			Cstr file_info(pNode->m_strName);
			if (ffind.FindFile(pNode->m_strPath)){
				CTime t;
				time_t tt;
				ffind.FindNextFile();
				file_info+="  ---  ";
				file_info+=ffind.GetLength();
				file_info+=" bytes  ---  ";
				ffind.GetCreationTime(t);
				tt=t.GetTime();
				file_info+=ctime(&tt);
			}
			gMaindialog->SetStatusTextAndIcon(file_info.buf,gImgList->ExtractIcon(pNode->m_icon));
		}
	}
	CListCtrl::OnMouseMove(nFlags, point);
}



