// FilerTree.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CFilerTree





CFilerTree::CFilerTree()
{
	hRoot=0;
	checkdirty=false;
	m_bDragging=false;
	m_pDragImage = NULL;
	m_nTimerID=0;
	m_hcurDrag = ::LoadCursor(AfxGetResourceHandle(),MAKEINTRESOURCE(IDC_DRAG));
    m_hcurDragCopy = ::LoadCursor(AfxGetResourceHandle(),MAKEINTRESOURCE(IDC_DRAGCOPY));
    m_hcurNodrop = ::LoadCursor(AfxGetResourceHandle(),MAKEINTRESOURCE(IDC_NODROP));
	itextbuff="";
}

CFilerTree::~CFilerTree()
{
	if(m_pDragImage) {
		delete m_pDragImage;
		m_pDragImage = NULL;
	}
}

char * CFilerTree::GetItText(HTREEITEM h,char * txt)
{
	TVITEM item;
	item.hItem = h;
	item.mask = TVIF_TEXT;
	int nLen = 128;
	Cstr str(nLen);
	int nRes;
	do
	{
		nLen *= 2;
		str.resize(nLen);
		str.len=nLen;
		item.pszText = str.buf;
		item.cchTextMax = nLen;
		::SendMessage(m_hWnd, TVM_GETITEM, 0, (LPARAM)&item);
		nRes = lstrlen(item.pszText);
	} while (nRes == nLen-1);
	if (txt!=NULL) {
		lstrcpy(txt,str.buf);
		return txt;
	} else {
		itextbuff=str.buf;
		return itextbuff.buf;
	}
}

void CFilerTree::DeleteAllItems()
{
	
	if (checkdirty)
		SaveIfDirty();
	if (hRoot!=0) {
		Expand(hRoot, TVE_COLLAPSE);
		CleanItemData(hRoot,false);
	}
	CTreeCtrl::DeleteAllItems();
}


void CFilerTree::CleanItemData(HTREEITEM h,bool del)
{
	HTREEITEM cItem=GetChildItem(h);		
	while (cItem!=NULL){				
		HTREEITEM cItemnext=GetNextSiblingItem(cItem);
		CleanItemData(cItem);
		cItem=cItemnext;
	}
	cleanOneItemData(GetItemData(h));
	if (del) DeleteItem(h);
}


HTREEITEM CFilerTree::GetSelected()
{
	hSelected=GetSelectedItem();
	if (hSelected!=NULL) {
		hSelectedType=GetItemType(hSelected);
		GetItText(hSelected,hSelectedText);
		GetItemPath(hSelected,hSelectedPath);
	}
	return hSelected;
}

void CFilerTree::SaveIfDirty()
{
	HTREEITEM h=GetChildItem(hRoot);
	while (h!=NULL){
		if (isDirty(h))
			doSave(h);
		h=GetNextSiblingItem(h);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CFilerTree message handlers

BEGIN_MESSAGE_MAP(CFilerTree,CTreeCtrl)
	//{{AFX_MSG_MAP(CFilerTree)
	ON_NOTIFY_REFLECT(TVN_SELCHANGED, OnSelchanged)
	ON_NOTIFY_REFLECT(TVN_ITEMEXPANDING, OnItemexpanding)
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	ON_NOTIFY_REFLECT(TVN_BEGINDRAG, OnBegindrag)
	ON_WM_TIMER()
	ON_NOTIFY_REFLECT(TVN_ENDLABELEDIT, OnEndlabeledit)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnDblclk)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


void CFilerTree::SetDropMark(CPoint pt)
{
	HTREEITEM olddropitem=m_hDropItem;
	if (WindowFromPoint (pt) != (CWnd *) this) {
		m_hDropItem = NULL;
	} else {			
		UINT uFlags;
		CPoint point(pt);
		ScreenToClient (&pt);
		m_hDropItem = HitTest (pt, &uFlags);
		
		if (m_hDropItem!=NULL && olddropitem!=m_hDropItem) {
			SetInsertMark(NULL);
			SelectDropTarget (NULL);

			EnsureVisible(GetPrevVisibleItem(m_hDropItem));
			EnsureVisible(GetNextVisibleItem(m_hDropItem));
			
			if (mayDropAfter(m_hDropItem)) {
				SetInsertMark(m_hDropItem,true);
				SelectDropTarget (GetParentItem(m_hDropItem));
				TimerOff();
			} else if (mayDropInto(m_hDropItem)) {
				SetInsertMark(GetChildItem(m_hDropItem),false);
				SelectDropTarget (m_hDropItem);
				TimerOn(pt);
			} else 
				m_hDropItem = NULL;
		}
	}

	if (m_hDropItem==NULL) {
			SetInsertMark(NULL);
			SelectDropTarget (NULL);
			TimerOff();
			SetCursor(m_hcurNodrop);
	} else {
		if (GetAsyncKeyState(VK_SHIFT) & 0x8000 || GetAsyncKeyState(VK_CONTROL) & 0x8000)
			SetCursor(m_hcurDragCopy);
		else
			SetCursor(m_hcurDrag);
	}
}


void CFilerTree::OnMouseMove(UINT nFlags, CPoint point) 
{
	if (m_bDragging) {
		CPoint pt (point);
		ClientToScreen (&pt);
		m_pDragImage->DragMove (pt);
		m_pDragImage->DragShowNolock (FALSE);
		SetDropMark(pt);
		m_pDragImage->DragShowNolock (TRUE);
	}	
	CTreeCtrl::OnMouseMove(nFlags, point);
}



void CFilerTree::DropItem()
{
	SelectDropTarget (NULL);
	SetInsertMark(NULL);
	if (m_hDropItem) {
		doDropItem();
	}
}



void CFilerTree::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (m_bDragging) {		
		SetCursor(m_hcurOld);
		m_pDragImage->DragLeave (GetDesktopWindow ());	
		m_pDragImage->EndDrag ();
		::ReleaseCapture ();
		m_bDragging = FALSE;

		CPoint pt (point);
		ClientToScreen (&pt);
		TimerOff();
		if (WindowFromPoint (pt)==(CWnd *) this)
			DropItem ();
	}
	CTreeCtrl::OnLButtonUp(nFlags, point);
}

void CFilerTree::OnBegindrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
	m_hDragItem = ((NM_TREEVIEW *)pNMHDR)->itemNew.hItem;

	if(m_pDragImage) {
		delete m_pDragImage;
		m_pDragImage=NULL;
	}
	
    m_hcurOld = ::SetCursor(NULL);
    ::SetCursor(m_hcurOld);
    

	if (mayDrag(m_hDragItem)) {    
		m_pDragImage = CreateDragImage (m_hDragItem);

		m_pDragImage->BeginDrag (0, CPoint (8, 8));
		m_pDragImage->DragEnter (GetDesktopWindow (), ((NM_TREEVIEW *)pNMHDR)->ptDrag);
		m_bDragging = TRUE;
		m_hDropItem = NULL;
		SetCapture ();
	}
	*pResult = 0;
}


int CFilerTree::OnContextMenu(CPoint point) 
{
	CPoint pt(point.x,point.y);
	ScreenToClient(&pt);
	HTREEITEM selitem=HitTest(pt);
	int kind=-1;
	if (selitem!=0) {
		kind = GetItemType(selitem);
		Select(selitem, TVGN_CARET);
	}
	return kind;
}

void CFilerTree::OnItemexpanding(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	HTREEITEM hItem=pNMTreeView->itemNew.hItem;
	*pResult = 0;
	if (hItem==NULL)return;
	if (pNMTreeView->action != TVE_EXPAND) return; //It has been explored 
	doOnExpand(hItem);
}

void CFilerTree::OnSelchanged(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	*pResult = 0;

	doOnSelchanged(pNMTreeView->itemOld.hItem,pNMTreeView->itemNew.hItem);
}

void CFilerTree::OnTimer(UINT nIDEvent) 
{
	if( nIDEvent != m_nTimerID ) {
		CTreeCtrl::OnTimer(nIDEvent);
		return;
	}

	TimerOff();

	HTREEITEM	trItem	= 0;
	UINT		uFlag	= 0;

	trItem = HitTest(m_TimerPoint, &uFlag);

	if ( trItem ) {
		if (mayDropInto(trItem)) {
			m_pDragImage->DragShowNolock (FALSE);
			HTREEITEM firstvisible=GetFirstVisibleItem();
			Expand(trItem,TVE_EXPAND);	
			SelectSetFirstVisible(firstvisible);
			m_pDragImage->DragShowNolock (TRUE);
		}
	}
}


void CFilerTree::TimerOff()
{
	if (m_nTimerID!=0) {
		KillTimer(  m_nTimerID );
		m_nTimerID = 0;
	}
}

void CFilerTree::TimerOn(CPoint pt)
{
	TimerOff();
	m_nTimerID = SetTimer(1, 500, NULL);
	m_TimerPoint = pt;
}

///////////////////////////////////////////////////////////////////////
int CFilerTree::FindTreeItem(HTREEITEM item,char * tofind,bool recurse)
{
	HTREEITEM i=GetChildItem(item);
	int hret=0;
	while (i!=NULL) {
		if (!strcmp(GetItText(i),tofind)) {
			hret=(int)i;
			break;
		}
		if (recurse) {
			if (ItemHasChildren(i)){
				hret=FindTreeItem(i,tofind,recurse);
				if (hret!=0) 
					break;
			}
		}
		i=GetNextSiblingItem(i);
	}
	return hret;
}



void CFilerTree::TreeToIni(HTREEITEM h,HTREEITEM hsection,int section)
{
	if (h!=NULL) {
		char * txt=GetItText(h);
		char * equal=strchr(txt,'=');
		if (*txt!='\0' && equal!=NULL) {
			*equal='\0';
			SetIniString(section,txt,equal+1);
		}
	} else {
		Cstr inibuff;
		HTREEITEM hitem=GetChildItem(hsection);
		int xtend=0;
		hitem=GetChildItem(hsection);
		while (hitem) {
			char *txt=GetItText(hitem);
			if (strcmp(txt,NEWITEMNAME) && *txt!='\0') {
				inibuff.addtoend(txt,xtend);
				xtend=1;
			}
			HTREEITEM hthis=hitem;
			hitem=GetNextSiblingItem(hitem);
			if (*txt=='\0')
				DeleteItem(hthis);
		}
		inibuff.addtoend("",xtend);
		WriteIniSection(section,inibuff.buf);
	}
}


void CFilerTree::OnEndlabeledit(NMHDR* pNMHDR, LRESULT* pResult) 
{
	TV_DISPINFO* pTVDispInfo = (TV_DISPINFO*)pNMHDR;
	*pResult = 0;
	doOnEndlabeledit(GetSelectedItem(),pTVDispInfo->item.pszText);
}

void CFilerTree::OnDblclk(NMHDR* pNMHDR, LRESULT* pResult) 
{
	if (!GetSelected()) return;
	doOnDblclk(hSelected);
	*pResult = 0;
}
