// FileTreeCtrl.cpp : implementation file

#include "stdafx.h"
#include "filerx.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFilerDlg * gMaindialog;
extern CImgList * gImgList;
/////////////////////////////////////////////////////////////////////////////
// CFileTreeCtrl

CFileTreeCtrl::CFileTreeCtrl()
{
	tree_type=0;
	m_pRoot=0;
	selNode=0;
}

CFileTreeCtrl::~CFileTreeCtrl()
{
	if (m_pRoot!=0)
		delete m_pRoot;
}

void CFileTreeCtrl::cleanOneItemData(UINT pData)
{
	((TreeNode*)pData)->Cleanup();
}

HTREEITEM CFileTreeCtrl::FindItem(const char* path)
{
	HTREEITEM  hItem = hItem = GetRootItem();
	TreeNode*   pNode = (TreeNode*)GetItemData(hItem);
	pNode=pNode->FindRootChild(path); //My Computer or Network ?
	hItem = pNode->m_hItem;

	char pth[MAX_PATH];
	char * src=(char *)path;
	char * dst=pth;
	bool done=FALSE;


	if (*src=='\\' && *(src+1)=='\\') {*dst++='\\';*dst++='\\';} //network ..
	while (!done) {
		Expand(hItem, TVE_EXPAND);						//Expand Item
		while(*src!='\\' && *src!='\0') *dst++=*src++;	//Reads Subfolder/file name
		if (*src=='\0') done=TRUE;						//Is it the end ?
		*dst='\0';										//to end string

		pNode = pNode->FindChild(pth);					//Find corresponding subnode
		if (pNode) 
			hItem = pNode->m_hItem;						//Found ...
		else 
			hItem=0;									//Not found ...
		if (done || !pNode) break;						
		*dst++=*src++;									//Copy slash
	}
	return hItem;
}

CFileTreeCtrl * thistree;

UINT ExploreThread( LPVOID pParam )
{	
	thistree->Explore(thistree->m_pRoot->m_hItem);
	gMaindialog->TryToShowPath();
    return 0;   // thread completed successfully
}

void CFileTreeCtrl::DisplayTree()
{
	m_pRoot = new TreeNode; //Build root node
	UINT nMask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_CHILDREN;
	m_pRoot->m_hItem = InsertItem(nMask, 
						m_pRoot->m_strName,m_pRoot->m_icon,m_pRoot->m_oicon, 
						NULL, NULL,(LPARAM)m_pRoot, TVI_ROOT, TVI_SORT);

	thistree=this;
	hRoot=m_pRoot->m_hItem;
	AfxBeginThread(ExploreThread, m_pRoot->m_hItem);
	//Explore(m_pRoot->m_hItem);
}


void CFileTreeCtrl::Explore(HTREEITEM hParent)
{
	UINT nMask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	TreeNode* pNode = (TreeNode*)GetItemData(hParent);
	if (pNode->m_explored) return;
	HCURSOR oldcur=SetCursor(LoadCursor(NULL,IDC_WAIT));
	//Build list of subnodes, Folders and Files
	pNode->EnumChildren();	
	//Explore Folders list only
	TreeNode * pChild=pNode->m_Folders;
	while (pChild!=NULL) {	
		if (pChild->m_hassubfolders) 
			nMask |= TVIF_CHILDREN;				
		else
			nMask &= ~TVIF_CHILDREN;
		pChild->m_hItem = InsertItem(nMask, 
						pChild->m_strName,pChild->m_icon,pChild->m_oicon, 
						NULL, NULL, (LPARAM)pChild, hParent, TVI_SORT);
		pChild=pChild->m_next;
	}
	SetCursor(oldcur);
}

void CFileTreeCtrl::GetSelectedNodePath(char *path)
{
	*path='\0';
	if (selNode!=NULL)
		lstrcpy(path,selNode->m_strPath);
}

void CFileTreeCtrl::Refresh(const char * pth)
{
	HTREEITEM hItem=NULL;
	char pthcopy[MAX_PATH];
	lstrcpy(pthcopy,pth);

	//Find the tree Item
	hItem = FindItem(pth);
	if (!hItem){
		char * slash=strrchr(pthcopy,'\\');	// find last slash
		if (slash){
			*slash='\0';
			hItem = FindItem(pthcopy);
		}
	}

	if (hItem){
		Select(GetParentItem(hItem), TVGN_CARET);//to force redraw
		HTREEITEM cItem=GetChildItem(hItem);		
		while (cItem!=NULL){					
			HTREEITEM cItemnext=GetNextSiblingItem(cItem);
			CleanItemData(cItem);
			cItem=cItemnext;
		}
		cleanOneItemData(GetItemData(hItem));
		Explore(hItem);
		SelectSetFirstVisible(hItem);
		Select(hItem, TVGN_CARET); //to redraw list
	}
}

/////////////////////////////////////////////////////////////////////////////
// CFileTreeCtrl message handlers


void CFileTreeCtrl::doOnSelchanged(HTREEITEM olditem,HTREEITEM newitem) 
{
	selNode=NULL;
	if (newitem==NULL) return;
	//Any new project file will be created in the selected folder
	selNode = (TreeNode*) GetItemData(newitem);
	//If node has not been explored before, must explore it
	Explore(newitem);
	//Show the file list
	gMaindialog->m_list.DisplayList(selNode);
	gMaindialog->SetStatusTextAndIcon(selNode->m_strPath , gImgList->ExtractIcon(selNode->m_icon));
}

void CFileTreeCtrl::doOnExpand(HTREEITEM hItem) 
{
	Explore(hItem);	
}

