// MacTreeCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"
#include "shortcutdlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFilerDlg * gMaindialog;
extern CSciCtrl * gScintilla;
//extern bool gPlaying;
extern bool gRecording;
extern bool gTesting;
extern char * gMyDirectory;
extern macro * gCurmac;
extern macro * gMacros;
/////////////////////////////////////////////////////////////////////////////
// CMacTreeCtrl

CMacTreeCtrl::CMacTreeCtrl()
{
	tree_type=3;
	newmacros=NULL;
	nodirtynotif=false;
	gMacros=NULL;
	gCurmac=NULL;
	selchanged=false;
}

CMacTreeCtrl::~CMacTreeCtrl()
{
}

void CMacTreeCtrl::InitTree()
{
	hRoot=InsertMacItem("MACROS FILES",macro_root,TVI_ROOT,TVI_LAST,TYPE_ROOT,false);
	LoadMacros();
	Expand(hRoot, TVE_EXPAND);

}

void CMacTreeCtrl::CleanupMacros()
{
	macro * m=gMacros;
	macro * mm;
	while (m!=NULL){
		mm=m;
		m=m->next;
		DeleteMacro(mm);
	}
}

void CMacTreeCtrl::SetDirty(HTREEITEM h)
{
	if (h==NULL) return;
	HTREEITEM hh=h;
	while(!isRootItem(hh)){
		if (!isDirty(hh)) {
			((pitemstruct *)GetItemData(hh))->dirty=true;
			SetMacroIcon(hh,true);
		}
		hh=GetParentItem(hh);
	}
}

void CMacTreeCtrl::NotifDirty()
{
	if (nodirtynotif) return;
	HTREEITEM h=GetSelectedItem();
	if (h!=NULL)
		if (isMacroItem(h))
			SetDirty(h);
}


HTREEITEM CMacTreeCtrl::InsertMacItem(const char * name,int image,HTREEITEM hParent,HTREEITEM after,int type,bool dirty)
{
	HTREEITEM newh=InsertItem(name,image,image,hParent,after);
	pitemstruct * pis=new(pitemstruct);
	pis->dirty=dirty;
	pis->type=type;
	SetItemData(newh,(UINT)pis);
	return newh;
}

HTREEITEM CMacTreeCtrl::InsertFile(const char * name,HTREEITEM parent)
{
	return InsertMacItem(name,macro_test,parent,TVI_LAST,TYPE_FILE,false);
}

HTREEITEM CMacTreeCtrl::InsertMacro(const char * name,HTREEITEM parent)
{
	return InsertMacItem(name,empty_macro, parent,TVI_LAST,TYPE_MACRO,false);
}

HTREEITEM CMacTreeCtrl::InsertMacroFile(const char * name)
{
	return InsertMacItem(name,macro_file,hRoot,TVI_LAST,TYPE_MACROFILE,false);
}

void CMacTreeCtrl::SetMacroIcon(HTREEITEM item, bool dirtyitem/*=false*/){

	int result;

	if (isMacroFileItem(item)) {
		result=(dirtyitem ? macro_file_dirty : macro_file );
	} else {
		if (dirtyitem) 
			result=macro_dirty;
		else  {
			macro * m=FindMacroByHandle(item);
			result = (m==NULL ? empty_macro : m->flags & 255);
		}
	}
	TVITEM tv;
	tv.hItem=item;
	tv.mask=TVIF_IMAGE |TVIF_SELECTEDIMAGE ;
	tv.iSelectedImage=result;
	tv.iImage=result;
	SetItem(&tv);
}

/////////////////////////////////////////////////////////////////////////////
// Load Macros


//From Ini File

void CMacTreeCtrl::LoadMacros()
{
	CIniSection * inisec=new CIniSection(SECTION_MACROS);
	while (inisec->GetNextKey()) {
		if (!strncmp(inisec->key,"testfile",8)){
			InsertFile(inisec->value,hRoot);
		}else{
			LoadMacroFile(inisec->value);
		}
	}
	delete inisec;
}

void CMacTreeCtrl::LoadMacroFile(const char * fname)
{
// Don't load file twice
	HTREEITEM macrofile=GetChildItem(hRoot);
	while (macrofile!=NULL){
		if (!strcmp(GetItText(macrofile),fname)) return;
		macrofile=GetNextSiblingItem(macrofile);	
	}
	Cstr buffer;
	ReadTextFile(fname,&buffer);
	if (buffer.len!=0) {
		HTREEITEM macrofile=InsertMacroFile(fname);
		TextToMacros(macrofile,NULL,buffer.buf);			
	}
}

/////////////////////////////////////////////////////////////////////////////
// Save Macros


void CMacTreeCtrl::doSave(HTREEITEM h)
{
	char * pname=GetItText(h);
	if (UserAccepts(IDS_SAVE_MACROS,pname)) {
		p.Open(pname,CFile::modeCreate | CFile::modeWrite);
		HTREEITEM macroitem=GetChildItem(h);		//Get First Macro
		while (macroitem!=NULL){					//got one ?
			char * words=0;
			macro * m=FindMacroByHandle(macroitem);
			words = Decompile(m,gMacros);
			if (words) {
				p.WriteString(words);
				delete[]words;
				p.WriteString("\r\n");
			}
			macroitem=GetNextSiblingItem(macroitem);
		}
		p.Close();
		((pitemstruct *)GetItemData(h))->dirty=false;
		SetMacroIcon(h);
	}
}

//////////////////////////////////////////////////////////////////////
// Folder Creation / Deletion

void CMacTreeCtrl::fnew()
{
	if (isMacroFileItem(hSelected)){
		TextToMacros(hSelected,NULL,"Macro void New_macro\n");
		SetDirty(hSelected);
		Expand(hSelected, TVE_EXPAND);
	}
}

void CMacTreeCtrl::fdel()
{	
	if (UserAccepts(IDS_IFERASE_MACRO,"")) {
		SetDirty(hSelected);
		CleanItemData(hSelected);
	}
}

//////////////////////////////////////////////////////////////////////
// MacroFile Creation / Deletion

void CMacTreeCtrl::pnew()
{
	char pname[MAX_PATH];
	lstrcpy(pname,gMyDirectory);
	lstrcat(pname,"new.macros");
	HTREEITEM hMacroFile = InsertMacroFile(pname);
	Expand(hRoot, TVE_EXPAND);
	Select(hMacroFile, TVGN_CARET);
}

void CMacTreeCtrl::DeleteMacroFile(HTREEITEM h)
{
	HTREEITEM macroitem=GetChildItem(h);		//Get First Macro
	while (macroitem!=NULL){					//got one ?
		DeleteMacroByName(GetItText(macroitem));
		macroitem=GetNextSiblingItem(macroitem);
	}
}

void CMacTreeCtrl::pdel()
{
	if (UserAccepts(IDS_IFCLOSE_MACROFILE,"")) {
		if (isDirty(hSelected))
			doSave(hSelected);
		DeleteMacroFile(hSelected);
		CleanItemData(hSelected);
	}
}

void CMacTreeCtrl::CompileEditedMacro(HTREEITEM citem/*=NULL*/)
{
	if (gMaindialog->viewmode2==VIEW2_SOURCE) {
		HTREEITEM item=citem;
		if (item==NULL)
			item=GetSelectedItem();

		if (item!=NULL) {
			if (isMacroItem(item) && isDirty(item)) {
				Cstr textbuffer;
				gScintilla->GetBuffer(&textbuffer);
				TextToMacros(GetParentItem(item),item,textbuffer.buf);				
				((pitemstruct *)GetItemData(item))->dirty=false;
				SetMacroIcon(item);
				selchanged=true;
			}
		}
	}
}

void CMacTreeCtrl::SynchronizeSource(HTREEITEM selitm)
{
	
	if (!selchanged) return;

	if (selitm!=NULL){
		if (isMacroItem(selitm)){
			gCurmac=FindMacroByHandle(selitm);
			DWORD hk=GetIniShortcutInt(gCurmac->name);
			gMaindialog->ShowHKName(gCurmac->shortcut);
		}
	
		if (gMaindialog->viewmode2==VIEW2_SOURCE) {
			selchanged=false;
			if (isMacroItem(selitm))
				ShowMacroText(selitm);
			else
				gScintilla->ClearAll();
		}	
	}
}

void CMacTreeCtrl::SetBold(HTREEITEM h)
{
	
	macro * m=FindMacroByHandle(h);
	if (m!=NULL) {
		if (m->shortcut!=0)
			SetItemState(h,TVIS_BOLD,TVIS_BOLD);
		else
			SetItemState(h,0,TVIS_BOLD);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CMacTreeCtrl message handlers

void CMacTreeCtrl::doOnDblclk(HTREEITEM selitm) 
{
	if (gRecording) return;
	
	char * selitemtext=GetItText(selitm);
	if (isMacroFileItem(selitm)) {
		if (!gTesting)
			gMaindialog->SciTEOpen(selitemtext,true);
	} else if (isMacroItem(selitm)){
		gMaindialog->UpdateViews(-1,VIEW2_SOURCE);
		//ShowMacroText(selitm);
	} else if (isFileItem(selitm)){
		if (!gTesting) {
			gMaindialog->EnterTestMode();
			gScintilla->OpenText(selitemtext);
		}
	}
}

void CMacTreeCtrl::doOnEndlabeledit(HTREEITEM sitem,char * newtext) 
{
	if (gRecording) return;
	
	if (newtext!=NULL){
		char newname[MAX_PATH];
		lstrcpy(newname,newtext);
		HTREEITEM sitem=GetSelectedItem();
		bool ismac=isMacroItem(sitem);
		if (ismac){
			macro * m=FindMacroByHandle(sitem);
			*(m->name)='\0';
			CheckMacroName(newname);
			lstrcpy(m->name,newname);
		}
		SetItemText(sitem,newname);
		SetDirty(sitem);
		if (ismac)
			ShowMacroText(sitem);
		gMaindialog->UpdateViews(-1,-1);
	}
}

void CMacTreeCtrl::doDropItem()
{
	HTREEITEM after;
	after=m_hDropItem;
	if (isMacroItem(m_hDropItem)) 
		m_hDropItem=GetParentItem(m_hDropItem);
	else if (isMacroFileItem(m_hDropItem))
		after=TVI_FIRST;
	else
		return;

	SetDirty(GetParentItem(m_hDragItem));
	HTREEITEM newh=InsertMacItem(GetItText(m_hDragItem),good_macro,m_hDropItem,after,TYPE_MACRO,true);
	Select(newh, TVGN_CARET);
	SetBold(newh);
	SetDirty(m_hDropItem);
	CleanItemData(m_hDragItem);
}

void CMacTreeCtrl::doOnExpand(HTREEITEM hItem) 
{
	HTREEITEM h=GetChildItem(hItem);
	while(h!=0) {
		SetBold(h);
		h=GetNextSiblingItem(h);
	}
}

void CMacTreeCtrl::doOnSelchanged(HTREEITEM olditem,HTREEITEM newitem) 
{
	if (gRecording) return;
	selchanged=true;
	CompileEditedMacro(olditem);
	SynchronizeSource(newitem);
	gMaindialog->UpdateViews(-1,-1);
}

///////////////////////////////////////////////////////////
// COMPILE/DECOMPILE

void CMacTreeCtrl::DoDecompile(macro * m)
{
	char * words=Decompile(m,gMacros);
	if (words){
		nodirtynotif=true;
		gScintilla->SetText(words);
		nodirtynotif=false;
		delete[]words;
	}
}

void CMacTreeCtrl::ShowMacroText(HTREEITEM h)
{
	gCurmac=FindMacroByHandle(h);
	if (gMaindialog->viewmode2==VIEW2_SOURCE) {
		DoDecompile(gCurmac);
	}
}

void CMacTreeCtrl::DumpMacroCode()
{
	if (gMaindialog->viewmode2==VIEW2_SOURCE) {
		HTREEITEM h=GetSelectedItem();
		if (h) {
			macro * m=FindMacroByHandle(h);
			char * words=Dumpcode(m);
			if (words){
				nodirtynotif=true;
				gScintilla->SetText(words);
				delete[]words;
			}
		}
	}
}

void CMacTreeCtrl::TextToMacros(HTREEITEM macfile,HTREEITEM mac,char * buf)
{
	char * mbeg=buf;
	char * mend=buf;
	HTREEITEM macitem=mac;
	bool mustcreate=(mac==NULL);
	macro * newmacro;

	while (*mend!='\0'){
		
		newmacro=CreateMacro();
		Compile(newmacro,gMacros,mbeg,&mend);

		if (macitem!=NULL){
			gCurmac=FindMacroByName(newmacro->name);
			if (gCurmac==NULL)
				mustcreate=true;
		}
		if (mustcreate) {
			gCurmac=AppendMacroToList(newmacro);
			HTREEITEM hmac=InsertMacro(gCurmac->name,macfile);
			SetMacroIcon(hmac);

		} else {
			DeleteMacro(newmacro);
			Compile(gCurmac,gMacros,mbeg,&mend);
			SetMacroIcon(macitem);
		}
		mbeg=mend;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  DELETE/FIND/ADD

void CMacTreeCtrl::DeleteMacroByName(const char * name){
	macro * m=FindMacroByName(name);
	if (m!=NULL) {
		if (gCurmac==m) gCurmac=NULL;
		m->flags|=erased_macro;
		UnregisterHotKey(gMaindialog->m_hWnd,m->shortcut);
	}
}

macro * CMacTreeCtrl::FindMacroByHandle(HTREEITEM h)
{
	return FindMacroByName(GetItText(h));
}

macro * CMacTreeCtrl::FindMacroByName(const char * mname){
	macro * m=gMacros;
	while (m!=NULL){
		if ((m->flags & erased_macro)==0)
			if (!lstrcmp(m->name,mname)) 
				return m;
		m=m->next;
	}
	return NULL;
}

macro * CMacTreeCtrl::FindMacroByShortcut(int shortcut){
	macro * m=gMacros;
	while (m!=NULL){
		if ((m->flags & erased_macro)==0)
			if (m->shortcut==shortcut) 
				return m;
		m=m->next;
	}
	return NULL;
}

void CMacTreeCtrl::CheckMacroName(char * nam){	
	char * mmm=strstr(nam,"Macro");
	if (mmm!=NULL) *mmm='m';
	int nlen=lstrlen(nam);
	int xnum=0;
	while (FindMacroByName(nam)!=NULL){
		xnum++;
		wsprintf(nam+nlen,"_%u",xnum);
	}
}

macro * CMacTreeCtrl::AppendMacroToList(macro * m,char * name/*=NULL*/){
	char nbuff[100];
	char * nname=( name!=NULL ? name : m->name);
	strncpy(nbuff,nname,96);
	CheckMacroName(nbuff);
	lstrcpy(m->name,nbuff);

	macro *mm=gMacros;
	if (mm!=NULL){
		while (mm->next!=NULL) mm=mm->next;
		m->prev=mm;
		mm->next=m;
	}else
		gMacros=m;

	return m;
}

void CMacTreeCtrl::AddRecordedMacro()
{
	HTREEITEM destfile=GetSelected();
	if (destfile!=NULL) {
		if (isMacroItem(destfile)) 
			destfile=GetParentItem(destfile);
		else if (isFileItem(destfile) || isRootItem(destfile)) 
			destfile=NULL;
	}
	if (destfile==NULL) {
		if (newmacros==NULL)
			newmacros=InsertMacroFile("recorded.macros");
		destfile=newmacros;
	}
	AppendMacroToList(gCurmac,"Recorded_macro");
	HTREEITEM hmacro=InsertMacro(gCurmac->name,destfile);
	SetDirty(destfile);
	SetMacroIcon(hmacro);
	Select(hmacro, TVGN_CARET);
}

