// ProjTreeCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "SciCtrl.h"
#include "inifile.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"
#include  <io.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


extern	int folderIcon;				//closed folder icon
extern int ofolderIcon;			//open folder icon
extern CFilerDlg * gMaindialog;
extern CImgList * gImgList;

extern bool gPlaying;

/////////////////////////////////////////////////////////////////////////////
// CProjTreeCtrl

CProjTreeCtrl::CProjTreeCtrl()
{
	tree_type=1;
	//pIcon=0;
}

CProjTreeCtrl::~CProjTreeCtrl()
{
}


void CProjTreeCtrl::cleanOneItemData(UINT pData)
{
	delete ((pitemstruct *)pData);
}


HTREEITEM CProjTreeCtrl::InsertProjItem(const char * name,int image,int simage,HTREEITEM hParent,HTREEITEM after,int type,bool dirty)
{
	HTREEITEM newh=InsertItem(name,image,simage,hParent,after);
	pitemstruct * pis=new(pitemstruct);
	pis->dirty=dirty;
	pis->type=type;
	SetItemData(newh,(UINT)pis);
	return newh;
}


HTREEITEM CProjTreeCtrl::InsertFile(const char * name,HTREEITEM parent,HTREEITEM after)
{
	int iIcon = GetIcon(name);
	HTREEITEM newh=InsertProjItem(name,iIcon,iIcon,parent,after,TYPE_FILE,false);
	CFileFind ffind;
	if (ffind.FindFile(name)) {
		ffind.FindNextFile();
		InsertProjItem(name,iIcon,iIcon,newh,TVI_LAST,0,false);
		SetItemText(newh,ffind.GetFileName());
	}
	return newh;
}

HTREEITEM CProjTreeCtrl::InsertFolder(const char * name,HTREEITEM parent)
{
	return InsertProjItem(name,folderIcon, ofolderIcon,parent,TVI_LAST,TYPE_FOLDER,false);
}

HTREEITEM CProjTreeCtrl::InsertProject(const char * name)
{
	return InsertProjItem(name,piconindex, piconindex,hRoot,TVI_LAST,TYPE_PROJECT,false);
}



void CProjTreeCtrl::InitTree()
{
	HICON hicon=(HICON)FlxLoadIcon(IDI_ICON_RECORD);
	infoiconindex=gImgList->AddImage(hicon,INFOICON);
	hicon=(HICON)FlxLoadIcon(IDI_ICON_NEWPROJECT);
	ppiconindex=gImgList->AddImage(hicon,PPICON);
	hicon=(HICON)FlxLoadIcon(IDI_ICON_NEWMACROFILE);
	piconindex=gImgList->AddImage(hicon,PICON);

	hRoot=InsertProjItem("PROJECT FILES",ppiconindex,ppiconindex,TVI_ROOT,TVI_LAST,TYPE_ROOT,false);	
	LoadProjects();
}

void CProjTreeCtrl::LoadProjects()
{
	int pcount=0;
	SetRedraw(false);
	CIniSection * inisec=new CIniSection(SECTION_PROJECTS);
	while (inisec->GetNextKey()) {
		if (inisec->value!='\0') {
			pcount++;
			Cstr ppj("Project ");
			ppj+=pcount;
			if (gMaindialog->m_splash!=NULL && !gPlaying)
				gMaindialog->m_splash->SetMessage(ppj.buf);
			AddProject(inisec->value);
		}
	}
	delete inisec;
	SetRedraw(true);
}



void CProjTreeCtrl::SetDirty(HTREEITEM h)
{
	if (h==NULL) return;
	HTREEITEM hh=h;
	while(!isRootItem(hh)){
		if (isProjectItem(hh)) ((pitemstruct *)GetItemData(hh))->dirty=true;
		hh=GetParentItem(hh);
	}
}


int CProjTreeCtrl::GetIcon(const char * s)
{
	int ic=0;
	SHFILEINFO  shfi;
	if (SHGetFileInfo(s,0,&shfi,sizeof(shfi),SHGFI_ICON |SHGFI_SMALLICON | SHGFI_SYSICONINDEX))
		ic=shfi.iIcon;
	return ic;
}


/////////////////////////////////////////////////////////////////////////////
// Open Projects



bool CProjTreeCtrl::AddProject(const char * pname)
{
	if ((_access( pname, 0 ) == -1) || (*pname=='\0')) return false;
	if (!IsProject(pname)) return false;
	HTREEITEM hProject = InsertProject(pname);	//Root Child : Project
	p.Open(pname,CFile::modeRead);								//Open project file
	AddNewBranch("",hProject);									//Read it
	p.Close();													//And close
	Expand(hRoot, TVE_EXPAND);
	return true;
}


bool CProjTreeCtrl::ReadLine(Cstr * line)
{
	char rbuff[257];
	if (p.ReadString(rbuff,256)==NULL) return false;
	*line=rbuff;
	while (line->buf[line->len-1]!='\n') {
		if (p.ReadString(rbuff,256)==NULL) break;
		(*line)+=rbuff;
	}
	return true;
}

void CProjTreeCtrl::AddNewBranch(const char * s,HTREEITEM h)
{
	Cstr line(256);
	HTREEITEM newh=h;		
	if (*s!='\0')				// When called by AddProject, s is empty : don't add item
		newh = InsertFolder(s,h);	//Insert folder
	
	while(1) {
		if (!ReadLine(&line)) return;
		line.trim();
		if (line.len!=0){							// skip if empty
			if (line.buf[0]==']') return;		// end of folder : done 
			if (line.buf[0]=='[') {				// new folder
				//Cstr line1=line.mid(1);
				AddNewBranch(line.mid(1).buf,newh);
			}
			else								// file
				InsertFile(line.buf,newh);	//Insert File
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
// Import .dsp Project
bool CProjTreeCtrl::ImportDSP(const char * pnam)
{
	Cstr pname(pnam);
	Cstr lcaseline,r2,r3,r4;
	Cstr basepath,actbasepath;
	Cstr line(256);

	HTREEITEM hParent,newh;
	//Create Project Root and 4 subfolders
	HTREEITEM hProject = InsertProject(pnam);
	SetDirty(hProject);
	HTREEITEM hsource = InsertFolder("Source",hProject);
	HTREEITEM hheader = InsertFolder("Header",hProject);
	HTREEITEM hres = InsertFolder("Resource",hProject);
	HTREEITEM hmisc = InsertFolder("Misc",hProject);

	//Extract directory of this .dsp file
	int slash=pname.len-1;
	while (pname[slash]!='\\') slash--;
	basepath=pname.left(slash+1);
	
	p.Open(pnam,CFile::modeRead);
	while (ReadLine(&line)){
		line.trim();
		if (line.len!=0){							// skip if empty
			lcaseline=line;
			CharLower(lcaseline.buf);
			if (lcaseline.left(7)=="source="){
				line=line.mid(7);
				//Kind of file ...
				r2=lcaseline.right(2);
				r3=lcaseline.right(3);
				r4=lcaseline.right(4);
				if ((r2==".c") || (r4==".cpp") || (r4==".cxx") || (r4==".tli"))
					hParent=hsource;
				else if  ((r2==".h")|| (r4==".hpp") || (r4==".hxx") || (r4==".tlh"))
					hParent=hheader;
				else if ((r3==".rc") || (r4==".rct"))
					hParent=hres;
				else
					hParent=hmisc;
				// Build file path
				actbasepath=basepath;
				if (line.left(2)==".\\")
					line=line.mid(2);		//Same directory as .dsp
				else
					while (line.left(3)=="..\\"){ 
						line=line.mid(3);						//Remove "..\"
						slash=actbasepath.len-1;
						while (actbasepath[slash]!='\\') slash--;
						actbasepath=actbasepath.left(slash+1);
					}
				actbasepath+=line;
				
				newh=InsertFile(actbasepath.buf,hParent);
			}
		}
	}
	
	p.Close();
	Expand(hRoot, TVE_EXPAND);
	return true;
}


//////////////////////////////////////////////////////////////////////////
// When pasting or double clicking in list

void CProjTreeCtrl::AddNewFile(const char * s,HTREEITEM after)
{
	if (GetSelected()){
		if (isProjectOrFolderItem(hSelected)){
			SetDirty(hSelected);
			HTREEITEM newh = InsertFile(s,hSelected,after); 
			Expand(hSelected, TVE_EXPAND);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// Save Projects




void CProjTreeCtrl::doSave(HTREEITEM h)
{
	char * pname=GetItText(h);

	if (UserAccepts(IDS_SAVE_PROJECT,pname)) {
		p.Open(pname,CFile::modeCreate | CFile::modeWrite);
		HTREEITEM cItem=GetChildItem(h);		//Get First Project Child
		while (cItem!=NULL){					//got one ?
			SaveBranch(cItem,"");				//Save it
			cItem=GetNextSiblingItem(cItem);	//And get next
		}
		p.Close();
		((pitemstruct *)GetItemData(h))->dirty=false;
	}
}


void CProjTreeCtrl::SaveBranch(HTREEITEM h,const char * indnt)
{
	Cstr line(indnt);
	if (isFolderItem(h)){
		//Save Folder Name
		line+="[";
		line+=GetItText(h);
		line+="\n";
		p.WriteString(line.buf);
		//Recurse into children
		HTREEITEM cItem=GetChildItem(h);
		while (cItem!=NULL){
			line=indnt;
			line+="\t";
			SaveBranch(cItem,line.buf);	//increase indent for subfolder
			cItem=GetNextSiblingItem(cItem);	
		}
		//End of Folder
		line=indnt;
		line+="]";
	} else if (isFileItem(h)) {
		//Save file path
		char ipath[MAX_PATH];
		GetItemPath(h,ipath);
		line+=ipath;
	} else
		line="";

	if (line.len!=0) {
		line+="\n";
		p.WriteString(line.buf);
	}

}

//////////////////////////////////////////////////////////////////////
// COPY



HTREEITEM CProjTreeCtrl::CopyItemTo(HTREEITEM h,HTREEITEM newparent,HTREEITEM after)
{
	HTREEITEM newh;			//copy of h
	HTREEITEM childitem=GetChildItem(h);	//h children
	HTREEITEM copy;
	int iIcon;
	int ioIcon;

	//any root child is a project
	bool convertToProject=(!isProjectItem(h)) && (isRootItem(newparent));
	//a project/folder has no project child :
	bool convertToFolder=(isProjectItem(h)) && (!isRootItem(newparent));

	int t=GetItemType(h);

	if ((isProjectItem(h) && !convertToFolder) || convertToProject){
		iIcon=piconindex;
		ioIcon=piconindex;
		t=TYPE_PROJECT;
	}else if ((isFolderItem(h) && !convertToProject) || convertToFolder){
		iIcon=folderIcon;
		ioIcon=ofolderIcon;
		t=TYPE_FOLDER;
	}else{
		if (childitem) {
			iIcon=GetIcon(GetItText(childitem));
		} else
			iIcon=GetIcon(GetItText(h));
		ioIcon=iIcon;
	}

	newh=InsertProjItem(GetItText(h),iIcon,ioIcon,newparent,after,t,true);
	
	childitem=GetChildItem(h);
	while (childitem!=NULL){
		copy=CopyItemTo(childitem,newh,TVI_LAST);
		childitem=GetNextSiblingItem(childitem);	
	}
	return newh;
}


char * CProjTreeCtrl::GetItemPath(HTREEITEM h, char * path)
{
	HTREEITEM hh=h;
	*path='\0';
	if (isFileItem(h) && ItemHasChildren(h))
		hh=GetChildItem(h);
	else if (!isProjectItem(h))
		hh=NULL;
	
	if (hh!=NULL)
		GetItText(hh,path);
	return path;
}


//////////////////////////////////////////////////////////////////////
// Folder Creation / Deletion


void CProjTreeCtrl::fnew()
{
	HTREEITEM newh = InsertFolder("New Folder",hSelected);
	Expand(hSelected, TVE_EXPAND);
	Select(newh, TVGN_CARET);
	SetDirty(newh);
}

void CProjTreeCtrl::fdel()
{
	if (UserAccepts(IDS_IFERASE_FOLDER,"")){
		SetDirty(hSelected);
		CleanItemData(hSelected);
	}
}


//////////////////////////////////////////////////////////////////////
// Project Creation / Deletion

void CProjTreeCtrl::pnew()
{
	char pname[MAX_PATH];
	gMaindialog->m_tree.GetSelectedNodePath((char *)pname);
	lstrcat(pname,"\\new.project");
	HTREEITEM hProject = InsertProject(pname);
	Expand(hRoot, TVE_EXPAND);
	Select(hProject, TVGN_CARET);
	SetDirty(hProject);
}

void CProjTreeCtrl::pdel()
{
	if (UserAccepts(IDS_IFCLOSE_PROJECT,"")){
		if (isDirty(hSelected))
			doSave(hSelected);
		CleanItemData(hSelected);
	}
}


///////////////////////////////////////////////////////////////////////
// 



void CProjTreeCtrl::doOnDblclk(HTREEITEM h) 
{
	if (hSelectedType==0) return;

	if (IsMacros(hSelectedPath)){
		gMaindialog->m_mactree.LoadMacroFile(hSelectedPath);
	}else if (IsLua(hSelectedPath)){
		gMaindialog->m_scintilla.LoadLuaFile(hSelectedPath);
	}else if (gMaindialog->viewmode==VIEW_PROJS){
		gMaindialog->SciTEOpen(hSelectedPath,true);
	}
}


void CProjTreeCtrl::doOnEndlabeledit(HTREEITEM eitem,char * newtext)
{
	if (newtext!=NULL) {
		int stype=GetItemType(eitem);
		if (stype!=0) {
			if (stype==TYPE_FILE) {
				HTREEITEM newh=InsertFile(newtext,GetParentItem(eitem),eitem);
				CleanItemData(eitem,true);
				Select(newh,TVGN_CARET);
				SetDirty(newh);
			} else {
				SetItemText(eitem,newtext);
				SetDirty(eitem);
			}
		}
	}
}

void CProjTreeCtrl::doOnSelchanged(HTREEITEM itemold, HTREEITEM itemnew) 
{	
	if (m_bDragging) return;
	gMaindialog->UpdateViews(-1,-1);
}

void CProjTreeCtrl::AddTimeChild(HTREEITEM h,CTime t,const char * s)
{
	char ttt[256];
	sprintf(ttt,"%s : %i/%i/%i - %i:%i:%i",s,t.GetYear(),t.GetMonth(),t.GetDay(),t.GetHour(),t.GetMinute(),t.GetSecond());
	InsertProjItem(ttt,infoiconindex,infoiconindex,h,TVI_LAST,0,false);
}

void CProjTreeCtrl::doOnExpand(HTREEITEM hItem) 
{
	if (GetItemType(hItem)!=TYPE_FILE) return;
	HTREEITEM hchild=GetChildItem(hItem);
	if (hchild==NULL) return;
	if (GetNextSiblingItem(hchild)!=NULL) return; // already completed
	char * name=GetItText(hchild);
	CFileFind ffind;
	CTime t;
	if (ffind.FindFile(name)){
		ffind.FindNextFile();
		HTREEITEM hcre;
		char length[25];
		sprintf(length,"length = %i bytes",ffind.GetLength());
		hcre=InsertItem(length,infoiconindex,infoiconindex,hItem);
		pitemstruct * pis=new(pitemstruct);
		pis->dirty=false;
		SetItemData(hcre,(UINT) pis);
		SetNullType(hcre);
		ffind.GetCreationTime(t);
		AddTimeChild(hItem,t,"Created");
		ffind.GetLastAccessTime(t);
		AddTimeChild(hItem,t,"Last access");
		ffind.GetLastWriteTime(t);
		AddTimeChild(hItem,t,"Last write");
	}
}



void CProjTreeCtrl::doDropItem()
{
	HTREEITEM after;
	if (isFileItem(m_hDropItem)) {
		after=m_hDropItem;
		m_hDropItem=GetParentItem(m_hDropItem);
	} else {
		after=TVI_FIRST;
	}
	if (isProjectOrFolderItem(m_hDropItem)) {
		if (m_hDropItem==m_hDragItem) return;
		HTREEITEM p=GetParentItem(m_hDropItem);
		while((p!=hRoot) && (p!=0)){
			if (p==m_hDragItem) return;		//may not copy item into itself
			p=GetParentItem(p);
		}

		if (m_hDropItem!=m_hDragItem) {
			HTREEITEM newitem=CopyItemTo(m_hDragItem,m_hDropItem,after);
			Select(newitem, TVGN_CARET);
			if (!(GetAsyncKeyState(VK_SHIFT) & 0x8000) && !(GetAsyncKeyState(VK_CONTROL) & 0x8000))
				DeleteItem (m_hDragItem);
			SetDirty(m_hDropItem);
		}
	}
}


void CProjTreeCtrl::DropItemFromList ()
{
	SelectDropTarget (NULL);
	SetInsertMark(NULL);
	if (m_hDropItem) {
		HTREEITEM after;
		if (isFileItem(m_hDropItem)) {
			after=m_hDropItem;
			m_hDropItem=GetParentItem(m_hDropItem);
		} else {
			after=TVI_FIRST;
		}

		if (gMaindialog->viewmode==VIEW_PROJS) {
			if (isProjectOrFolderItem(m_hDropItem)) {
				char selpath[MAX_PATH];
				gMaindialog->m_list.GetFirstSelectedPath((char *)selpath);
				HTREEITEM newh;
				while (*selpath!='\0') {
					newh = InsertFile(selpath,m_hDropItem,after); 
					after=newh;
					gMaindialog->m_list.GetNextSelectedPath((char *)selpath);
				}
				Select(newh, TVGN_CARET);
				SetDirty(m_hDropItem);
				Expand(m_hDropItem, TVE_EXPAND);
			}
		}
	}
}


void CProjTreeCtrl::ExecMenuCommand(int rep)
{
	GetSelected();
	if (rep>10000) {
		gMaindialog->RunMenuMacro(hSelectedPath);
		return;
	}

	switch (rep){
		case IDS_EDIT			: gMaindialog->SciTEOpen(hSelectedPath,true);break;
		case IDS_WINDOWS_OPEN	: gMaindialog->WindowsOpen(hSelectedPath);break;
		case IDS_SAVE		: doSave(hSelected);break;
		case IDS_DELPROJ	: pdel();break;
		case IDS_DELFOLDER	: fdel();break;
		case IDS_NEWFOLDER	: fnew();break;
		case IDS_DELETE		: CleanItemData(hSelected);break;
		case IDS_NEWPROJ	: pnew();break;
		case IDS_ADDTOINI	: SetIniNewString(SECTION_PROJECTS,hSelectedText);
								gMaindialog->m_proptree.ReloadIni();
								break;
		case IDS_SORT		: SortChildren(hSelected); break;
		case IDS_LOADSCRIPT :gMaindialog->ToSendMessage(SCD_LOADSCRIPT,hSelectedPath);
	}
}
