// PropTreeCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFilerDlg * gMaindialog;

//filerx.ini must be the last one
static char*  groupnames[] = { "dyn","local","user","abbrev","base","embed","Filerx.ini"};


#define TYPE_PROPROOT 0
#define TYPE_PROPGROUPFILERX 1
#define TYPE_PROPGROUPSCITE 2
#define TYPE_PROPSECTIONFILERX 3
#define TYPE_PROPSECTIONSCITE 4
#define TYPE_PROPKEY 5

/////////////////////////////////////////////////////////////////////////////
// CPropTreeCtrl

CPropTreeCtrl::CPropTreeCtrl()
{
	tree_type=2;
}

CPropTreeCtrl::~CPropTreeCtrl()
{
}

UINT PropsThread( LPVOID pParam )
{
	CPropTreeCtrl * tree=(CPropTreeCtrl *) pParam;
	for (int i=0;i<=GROUP_MAX-1;i++) {
		gMaindialog->ToSendMessage(SCD_ENUMPROPERTIES,groupnames[i]);
	}
	tree->Expand(tree->hRoot, TVE_EXPAND);
    return 0;   // thread completed successfully
}

void CPropTreeCtrl::InitTree()
{
	DeleteAllItems();
	hRoot = InsertItem("PROPERTIES",TYPE_PROPROOT,TYPE_PROPROOT);
	SetItemData(hRoot,TYPE_PROPROOT);

	for (int i=0;i<=GROUP_MAX;i++) {
		int typ=TYPE_PROPGROUPSCITE;
		if (i==GROUP_FILERX) //Filerx.ini
			typ=TYPE_PROPGROUPFILERX;
		hprops[i] = InsertItem(groupnames[i],typ, typ,hRoot);
		SetItemData(hprops[i],typ);
	}

	IniToTree();
	Expand(hprops[GROUP_FILERX],TVE_EXPAND);
	AfxBeginThread(PropsThread, LPVOID(this));
}

void CPropTreeCtrl::ReloadIni()
{
	if (gMaindialog->viewmode==VIEW_PROPS) {
		HTREEITEM hfilerx=hprops[GROUP_FILERX];
		HTREEITEM hsel=GetSelectedItem();
		if (hsel==NULL) hsel=hfilerx;
		int hseltyp=GetItemData(hsel);
		if (hseltyp==TYPE_PROPKEY) {
			hsel=GetParentItem(hsel);
			hseltyp=GetItemData(hsel);
		}
		char * htxt=GetItText(hsel);
		char * hselname=new char[lstrlen(htxt)+1];
		lstrcpy(hselname,htxt);
		HTREEITEM h=GetChildItem(hfilerx);
		while (h!=NULL) {
			HTREEITEM hnext=GetNextSiblingItem(h);
			DeleteItem(h);
			h=hnext;
		}
		
		IniToTree();
		
		Expand(hfilerx,TVE_EXPAND);
		h=GetChildItem(hfilerx);
		while (h!=NULL) {
			if (!strcmp(GetItText(h),hselname)) {
				SelectItem(h);
				Expand(h,TVE_EXPAND);
				break;
			}
			h=GetNextSiblingItem(h);
		}
		delete [] hselname;
	}
}

char * isprefx(char *target, const char *prefix) {
	while (*target && *prefix) {
		if (*target != *prefix)
			return NULL;
		target++;
		prefix++;
	}
	if (*prefix==0 && *target==':')
		return target+1;
	else
		return NULL;
}

void CPropTreeCtrl::AddProperty(char * p)
{
	HTREEITEM h=NULL;
	char * pdeb=NULL;

	for (int i=0;i<=GROUP_MAX-1;i++) {
		if (pdeb=isprefx(p,groupnames[i])) {
			h=hprops[i];
			break;
		}
	}
	if (!pdeb) return;

	char * pequal=strchr(pdeb,'=');
	char * ppoint=pdeb;
	bool found;
	HTREEITEM childitem;
	while(ppoint<pequal){
		ppoint=pdeb;
		while (ppoint<pequal){
			if (*ppoint=='.') break;
			if ((*ppoint=='$') && (*(ppoint+1)=='(')){
				while ((*ppoint)!=')') ppoint++;
			}
			ppoint++;
		}
		if (ppoint==pequal) break;
		*ppoint='\0';
		found=false;
		childitem=GetChildItem(h);
		while (childitem!=NULL){
			if (!strcmp(GetItText(childitem),pdeb)){
				h=childitem;
				found=true;
				break;
			}
			childitem=GetNextSiblingItem(childitem);	
		}
		if (!found){
			 h=InsertSection(pdeb,h);
		}
		pdeb=ppoint+1;
	}
	InsertKey(pdeb,h);
}


HTREEITEM CPropTreeCtrl::InsertKey(const char * name,HTREEITEM parent,HTREEITEM after)
{
	HTREEITEM newk=InsertItem(name,TYPE_PROPKEY,TYPE_PROPKEY,parent,after);
	SetItemData(newk,TYPE_PROPKEY);
	return newk;
}

HTREEITEM CPropTreeCtrl::InsertSection(const char * name,HTREEITEM parent)
{
	int typ=TYPE_PROPSECTIONSCITE;
	HTREEITEM after=TVI_SORT;
	if (GetItemData(parent)==TYPE_PROPGROUPFILERX) {
		typ=TYPE_PROPSECTIONFILERX;
		after=TVI_LAST;
	}
	
	HTREEITEM news = InsertItem(name,typ, typ,parent,after);
	SetItemData(news,typ);
	return news;
}

/////////////////////////////////////////////////////////////////////////////
// CPropTreeCtrl message handlers

void CPropTreeCtrl::doOnEndlabeledit(HTREEITEM h,char * newtext)
{
	Cstr ptosend;
	if (GetItemData(h)==TYPE_PROPKEY){
		HTREEITEM hsection=GetParentItem(h);
		if (GetParentItem(hsection)==hprops[GROUP_FILERX]) {
			if (newtext!=NULL) {
				if (!strcmp(GetItText(h),NEWITEMNAME))
					InsertKey(NEWITEMNAME,hsection,TVI_LAST);
				SetItemText(h,newtext);
				int snum=GetSectionNum(GetItText(hsection));
				if (snum==SECTION_MACROBAR)
					gMaindialog->RefreshMacroBar(hsection);
				else {
					if (snum!=SECTION_WINDOW && snum!=SECTION_WORK) 
						h=NULL; // Write whole section
					TreeToIni(h,hsection,snum);
				}
			}
		} else if (newtext!=NULL){
			if (strchr(newtext,'=')){
				SetItemText(h,newtext);
				ptosend=newtext;
				h=hsection;
				while (GetParentItem(h)!=hRoot){
					ptosend=Cstr(GetItText(h))+"."+ptosend;
					h=GetParentItem(h);
				}
				gMaindialog->ToSendMessage(SCD_SETPROPERTY,ptosend.buf);
			}
		}
	}
}


void CPropTreeCtrl::doDropItem()
{
	HTREEITEM after;
	after=m_hDropItem;
	m_hDropItem=GetParentItem(m_hDropItem);
	char * htxt=GetItText(m_hDragItem);
	HTREEITEM newitem=InsertKey(htxt,m_hDropItem,after);
	Select(newitem, TVGN_CARET);
	if (!(GetAsyncKeyState(VK_SHIFT) & 0x8000) && !(GetAsyncKeyState(VK_CONTROL) & 0x8000))
		DeleteItem (m_hDragItem);
		gMaindialog->RefreshMacroBar(m_hDropItem);
}

void CPropTreeCtrl::IniToTree()
{
	for (int i=0;i<=SECTION_MAX;i++) {
		HTREEITEM sectionitem=InsertSection(GetSectionName(i),hprops[GROUP_FILERX]);
		CIniSection * inisec=new CIniSection(i);
		inisec->separe=false;
		while (inisec->GetNextKey()){
			if (*inisec->key!=';')
				InsertKey(inisec->key,sectionitem,TVI_LAST);
		}
		InsertKey(NEWITEMNAME,sectionitem,TVI_LAST);
		delete inisec;
	}
}
