// PropTreeCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "inifile.h"
#include "SciCtrl.h"
#include "TreeCtrl_proj.h"
#include "ListCtrl.h"
#include "TreeCtrl_file.h"
#include "TreeCtrl_prop.h"
#include "TreeCtrl_mac.h"
#include "TreeCtrl_tlb.h"
#include "FilerDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFilerDlg * gMaindialog;
extern CSciCtrl * gScintilla;
extern bool gPlaying;
extern bool gTesting;

static char * g_names="Enum\0Struct\0Module\0Interface\0Dispinterface\0Coclass\0Typedef\0Union\0";

#define TYPE_TLBROOT 0
#define TYPE_REGFAV 1
#define TYPE_REG 2
#define TYPE_FILES 3

#define TYPE_TLB 4

#define TYPE_METHOD 5
#define TYPE_PROPERTYGET 6
#define TYPE_PROPERTYPUT 7
#define TYPE_PROPERTYPUTREF 8
#define TYPE_EVENT 9

#define TYPE_CONSTANT 10
#define TYPE_GROUPFOLDER 11

#define TYPE_EVENTINTERFACE 12
#define TYPE_ENUM 13
#define TYPE_STRUCT 14
#define TYPE_MODULE 15
#define TYPE_INTERFACE 16
#define TYPE_DISPINTERFACE 17
#define TYPE_COCLASS 18
#define TYPE_TYPEDEF 19
#define TYPE_UNION 20
#define TYPE_WHAT 21

#define TYPE_HELP 22
#define TYPE_CLSID 23
#define TYPE_OBJ 24
#define TYPE_TLBFILE 25


#define DONTFREE false

/////////////////////////////////////////////////////////////////////////////
// CTlbTreeCtrl

CTlbTreeCtrl::CTlbTreeCtrl()
{
	tree_type=4;
	m_pTypeLib=NULL;
	m_pCurrentTypeInfo=NULL;
}


CTlbTreeCtrl::~CTlbTreeCtrl()
{
	if (m_pCurrentTypeInfo) m_pCurrentTypeInfo->Release();
	if (m_pTypeLib) m_pTypeLib->Release();
}


void CTlbTreeCtrl::InitTree()
{
	hRoot = MyInsertItem("TYPELIBS",TYPE_TLBROOT,TVI_ROOT);
	hRegFav = MyInsertItem("Favorites",TYPE_REGFAV,hRoot);
	hReg = MyInsertItem("Registry",TYPE_REG,hRoot);
	hFiles = MyInsertItem("Files",TYPE_FILES,hRoot);
	Expand(hRoot, TVE_EXPAND);

	EnumerateRegistryForTypeLibEntries();
}


HTREEITEM CTlbTreeCtrl::MyInsertItem(const char * text,int typ,HTREEITEM parent,int index,bool withdummy)
{
	HTREEITEM newh=InsertItem(text,typ,typ,parent);	
	tmpdata = new CTypeLibraryData;
	tmpdata->m_path=NULL;
	tmpdata->type=typ;
	tmpdata->visited=false;
	tmpdata->index=index;
	SetItemData(newh,(unsigned long)tmpdata);
	if (withdummy) {
		HTREEITEM dummyh=InsertItem("empty",TYPE_WHAT,TYPE_WHAT,newh);
		CTypeLibraryData * pdata=new CTypeLibraryData;
		pdata->visited=false;
		pdata->type=TYPE_WHAT;
		pdata->m_path=NULL;
		SetItemData(dummyh,(unsigned long)pdata);
	}
	return newh;
}



void CTlbTreeCtrl::cleanOneItemData(UINT pData)
{
	CTypeLibraryData * p=(CTypeLibraryData *)pData;
	if (p->m_path!=NULL) delete [] p->m_path;
	delete p;
}



//////////////////////////////////////////////////////////////////////
// Fill "Registry" Tree Item


BOOL _RegOpenKey(HKEY key,LPCTSTR subkey,PHKEY result) {
	return (RegOpenKeyEx(key,subkey,0,KEY_READ,result)== ERROR_SUCCESS) ;
}

BOOL _RegEnumKey(HKEY key,DWORD index,CHAR * name) {
	return (RegEnumKey(key,index,name, MAX_PATH) == ERROR_SUCCESS);
}

DWORD _RegQuerySize(HKEY key) {
	long size;
	RegQueryValue(key,NULL,NULL,&size);
	return size;
}


BOOL CTlbTreeCtrl::EnumerateRegistryForTypeLibEntries()
{

	HKEY   hKeyTypeLib = NULL;
	HKEY   hKeyClassTLB = NULL;
	HKEY hKeyVersion = NULL;
	DWORD  dwIndex = 0;
	DWORD  dwCLSIDIndex = 0;
	long  dwSize = 0;
	CHAR  cCLSID[MAX_PATH+1];
	CHAR  cVersionInfo[MAX_PATH+1];
	HRESULT hr = S_OK;
	TCHAR* pDescription = NULL;

	CIniSection * inisec=new CIniSection(SECTION_TLBFAVORITES);
	//Open the TypeLib Key under HKCR
	if (_RegOpenKey(HKEY_CLASSES_ROOT,"TypeLib",&hKeyTypeLib)){
		//Enumerate CLSID entries
		while(_RegEnumKey(hKeyTypeLib,dwIndex,cCLSID)) {		
			if (_RegOpenKey(hKeyTypeLib,cCLSID, &hKeyClassTLB)) {       
 				dwCLSIDIndex = 0;	
				//Enumerate the CLSID subkeys
				while(_RegEnumKey(hKeyClassTLB,dwCLSIDIndex , cVersionInfo)) {					
					if (_RegOpenKey(hKeyClassTLB,cVersionInfo,&hKeyVersion)) {    					
						dwSize=_RegQuerySize(hKeyVersion);
						Cstr strDescription(dwSize);
						if (RegQueryValue(hKeyVersion,NULL,strDescription.buf,&dwSize)==ERROR_SUCCESS) {						
							strDescription.len=dwSize-1;
							strDescription+= " ";
							strDescription+=cVersionInfo;
							HTREEITEM hDest;
							if (inisec->KeyExists(strDescription.buf))
								hDest=hRegFav;
							else 
								hDest=hReg;
							HTREEITEM newh=MyInsertItem(strDescription.buf,TYPE_TLB,hDest);
							tmpdata->m_strDescription = strDescription;
							tmpdata->m_strCLSIDTypeLib = cCLSID;
							tmpdata->m_lMajorVersion = atoi(cVersionInfo);
							char* pMinor = strchr(cVersionInfo,'.');
							tmpdata->m_lMinorVersion = 0;
							if(pMinor) {	
								tmpdata->m_lMinorVersion = atoi(pMinor);
							}
							tmpdata->m_path=GetTlbPath(hKeyVersion);
							MyInsertItem(tmpdata->m_path,TYPE_TLBFILE,newh);
						}
						delete [] pDescription;
					}
					dwCLSIDIndex++;
				}
				RegCloseKey(hKeyClassTLB);
			}			
			dwIndex++;
		}
		RegCloseKey(hKeyTypeLib) ;
	}
	SortChildren(hReg);
	delete inisec;
	return TRUE;
}

char * CTlbTreeCtrl::GetTlbPath(HKEY key)
{
	CHAR tchLCID[10]; 
	char * path=new char[MAX_PATH+1];
	HKEY pathkey;
	HKEY keylcid;
	long pathsize=MAX_PATH;
	int index=0;
	bool bSuccess=false;

	*path='\0';
	while(_RegEnumKey(key,index,tchLCID)) {		
		if (_RegOpenKey(key,tchLCID,&keylcid)) {       
			if (_RegOpenKey(keylcid,"win32",&pathkey)) {    					
				if (RegQueryValue(pathkey,NULL,path,&pathsize)==ERROR_SUCCESS) {						
					bSuccess=true;
				}
				RegCloseKey(pathkey);
			}
			RegCloseKey(keylcid);
			if (bSuccess) 
				break;
		}
		index++;
	}
	return path;
}

BOOL CTlbTreeCtrl::GetInformationForTLB(HTREEITEM item, bool dovisit)
{
	BOOL bSuccess = TRUE;
	HRESULT hr = S_OK;
	BSTR bstrHelpString = NULL;
	BSTR bstrCLSID = NULL;
	BSTR bstrHelpFileName = NULL;
	CLSID clsidTypeLib;
	Cstr doc("");

	CTypeLibraryData* pData=(CTypeLibraryData*)GetItemData(item);
	bool isfile=(((CTypeLibraryData*)GetItemData(GetParentItem(item)))->type==TYPE_FILES);
	
	if(m_pTypeLib)
		m_pTypeLib->Release();
	
	if (isfile) {
			//Load directly from file
			bSuccess = LoadTypeLibFromFile(GetItemText(item),&m_pTypeLib);
	} else {
		BSTR bstrCLSID = pData->m_strCLSIDTypeLib.AllocSysString();
		doc=*BSTRtoCStr(bstrCLSID);
		if (CLSIDFromString(bstrCLSID,&clsidTypeLib)== S_OK) {
			//Try with a few lcids
			if (LoadRegTypeLib (clsidTypeLib,(unsigned short)pData->m_lMajorVersion,(unsigned short)pData->m_lMinorVersion,
								 0, &m_pTypeLib)!=S_OK)
				if (LoadRegTypeLib (clsidTypeLib,(unsigned short)pData->m_lMajorVersion,(unsigned short)pData->m_lMinorVersion,
								 0x409, &m_pTypeLib)!=S_OK)
					if (LoadRegTypeLib (clsidTypeLib,(unsigned short)pData->m_lMajorVersion,(unsigned short)pData->m_lMinorVersion,
								0x40C, &m_pTypeLib)!=S_OK) 
						bSuccess=false;
		} else {
			//Bad classid
			bSuccess=false;
		}
		if (bSuccess==false)
			//Try directly from file
			bSuccess =LoadTypeLibFromFile(pData->m_path,&m_pTypeLib);
	}
	

	if(bSuccess){			
		m_pTypeLib->GetDocumentation((unsigned int)-1, NULL, &bstrHelpString,NULL,&bstrHelpFileName);
		doc+="\n";
		doc+=*BSTRtoCStr(bstrHelpString);
		  			
		if(bstrHelpFileName && dovisit) {
			HTREEITEM newh=MyInsertItem(BSTRtoChar(bstrHelpFileName,false),TYPE_HELP,item);
		}
		doc +="\n";
		doc+=*BSTRtoCStr(bstrHelpFileName);

		
		if (!gPlaying) {
			gScintilla->ClearAllTlb();
			gScintilla->AddLine(doc.buf);
		}

		ExtractComponentsFromTLB(item,dovisit);		
	}
	return TRUE;
}

bool CTlbTreeCtrl::LoadTypeLibFromFile(const char * path,ITypeLib ** tlb)
{
	OLECHAR szTemp[512];
	MultiByteToWideChar(CP_ACP, 0,path, -1, szTemp, 512); 
	return (LoadTypeLib(szTemp,tlb)==S_OK) ;
}

void CTlbTreeCtrl::AddTypeLibFile(const char * path)
{
	ITypeLib * ptlib=NULL;
	if (LoadTypeLibFromFile(path,&ptlib)) {
		ptlib->Release();	
		HTREEITEM newh=MyInsertItem(path,TYPE_TLB,hFiles,-1,true);
		SelectItem(newh);
	}
}

BOOL  CTlbTreeCtrl::ExtractComponentsFromTLB(HTREEITEM item,bool dovisit)
{
	
	long lTypeInfoCount = 0;
	BSTR bstrName = NULL;
	
	if (dovisit) {
		lTypeInfoCount = m_pTypeLib->GetTypeInfoCount();
		if(lTypeInfoCount != 0){
			char * gnam=g_names;
			for (int i=0;i<8;i++) {
				groupfolderitem[i]=MyInsertItem(gnam,TYPE_GROUPFOLDER,item);
				gnam+=(lstrlen(gnam)+1);
			}

			TYPEKIND typekind;

			for(long lIter = 0; lIter  < lTypeInfoCount ; lIter++) {		  
				m_pTypeLib->GetDocumentation(lIter, &bstrName, NULL, NULL, NULL);
				m_pTypeLib->GetTypeInfoType(lIter, &typekind);
				HTREEITEM newh=MyInsertItem(BSTRtoChar(bstrName),typekind+TYPE_ENUM,groupfolderitem[typekind],lIter,true);
			}   
		
			for (i=0;i<8;i++) {
				if (!GetChildItem(groupfolderitem[i]))
					CleanItemData(groupfolderitem[i],true);
			}
		}
	}
	return TRUE;	
}

///////////////////////////////////////////////////////////////////////////////////
// Analyze one TypeInfo


BOOL CTlbTreeCtrl::GetTypeInfoInformation(HTREEITEM item, bool dovisit)
{
	TYPEKIND typeKind;
	BSTR bstrHelpFileName = NULL;
	BSTR bstrHelpString = NULL;
		
	CTypeLibraryData* pData=(CTypeLibraryData*)GetItemData(item);
	if (m_pTypeLib->GetTypeInfoType(pData->index, &typeKind)== S_OK) {
		if (!gPlaying) {
			if (m_pTypeLib->GetDocumentation((unsigned int)pData->index, NULL,&bstrHelpString, NULL,&bstrHelpFileName)==S_OK) {
				Cstr doc=*BSTRtoCStr(bstrHelpString);
				doc+="\nHelp File = ";
				doc+=*BSTRtoCStr(bstrHelpFileName);
				gScintilla->AddLine(doc.buf);
			}
		}
		if(m_pCurrentTypeInfo) 
			m_pCurrentTypeInfo->Release();		
		if (m_pTypeLib->GetTypeInfo((unsigned int)pData->index, &m_pCurrentTypeInfo)==S_OK) {
			GetTypeAttributes(item,dovisit);
			return true;
		}
	}
	
	return false;
}

void CTlbTreeCtrl::GetParameters(FUNCDESC* pFuncDesc,Cstr * rets)
{
	int pnum=pFuncDesc->cParams;
	unsigned short i;
	unsigned int cNames;
	const unsigned int MAX_NAMES = 40;
	BSTR rgNames[MAX_NAMES];

	m_pCurrentTypeInfo->GetNames(pFuncDesc->memid, rgNames, MAX_NAMES,&cNames);

	*rets="(";
	int deb=0;
	if (pFuncDesc->invkind==INVOKE_PROPERTYGET || pFuncDesc->invkind==INVOKE_FUNC) 
		deb++;
	if (pnum>0) {
		for(i = deb; i < cNames; i++){
			ELEMDESC param=pFuncDesc->lprgelemdescParam[i-deb];
			Cstr typorval("");
			GetType(&param.tdesc,&typorval);
			*rets+=(typorval+" ")+*BSTRtoCStr(rgNames[i],DONTFREE);
			if ((param.paramdesc.wParamFlags & (PARAMFLAG_FOPT | PARAMFLAG_FHASDEFAULT))== (PARAMFLAG_FOPT | PARAMFLAG_FHASDEFAULT)){
				GetValue(&param.paramdesc.pparamdescex->varDefaultValue,&typorval);
				*rets+="=";
				*rets+=typorval;
			}
			if (i<cNames-1)
				*rets+=", ";
		}
	}
	*rets+=")";
	for (i=0;i<cNames;i++)
		SysFreeString(rgNames[i]);
}

BOOL CTlbTreeCtrl::GetTypeAttributes(HTREEITEM item, bool dovisit)
{	
	MEMBERID memberID;
	FUNCDESC* pFuncDesc = NULL;
	VARDESC* pVarDesc = NULL;
	TYPEATTR* pTypeAttributes = NULL;
	BSTR bstrMethod = NULL;
	BSTR bstrProperty = NULL;
	BSTR bstrHelpString = NULL;
	HTREEITEM newh;
	
	if (!dovisit) return true;

	m_pCurrentTypeInfo->GetTypeAttr(&pTypeAttributes);
	
	CleanItemData(GetChildItem(item),true);//dummy
	//DeleteItem(GetChildItem(item)); //dummy
	if (pTypeAttributes->typekind==TKIND_ALIAS) {
		Cstr t("");
		GetType(&pTypeAttributes->tdescAlias,&t);
		newh=MyInsertItem(t.buf,TYPE_TYPEDEF,item);	
	}
	int iIndex = 0;
	for(int iIter = 0; iIter < pTypeAttributes->cFuncs; iIter++) {
		m_pCurrentTypeInfo->GetFuncDesc(iIter, &pFuncDesc) ;
		memberID = pFuncDesc->memid;
		m_pCurrentTypeInfo->GetDocumentation(memberID, &bstrMethod, &bstrHelpString, NULL, NULL);

		Cstr rettype;
		Cstr parameters;
		GetType(&pFuncDesc->elemdescFunc.tdesc,&rettype);
		GetParameters(pFuncDesc,&parameters);
		int typ;
		if (pFuncDesc->invkind==INVOKE_FUNC)
			typ=TYPE_METHOD;
		else if (pFuncDesc->invkind==INVOKE_PROPERTYGET)
			typ=TYPE_PROPERTYGET;
		else if (pFuncDesc->invkind==INVOKE_PROPERTYPUT)
			typ=TYPE_PROPERTYPUT;
		else if (pFuncDesc->invkind==INVOKE_PROPERTYPUTREF)
			typ=TYPE_PROPERTYPUTREF;
		else
			typ=TYPE_WHAT;
		Cstr txt=*BSTRtoCStr(bstrMethod);
		txt+=parameters;
		txt+=" : ";
		txt+=rettype;
		/*if ((pFuncDesc->elemdescFunc.paramdesc.wParamFlags & (PARAMFLAG_FHASDEFAULT))== (PARAMFLAG_FHASDEFAULT))
			txt=txt+"="+GetValue(&pFuncDesc->elemdescFunc.paramdesc.pparamdescex->varDefaultValue);*/

		newh=MyInsertItem(txt.buf,typ,item,iIter);
		if (bstrHelpString)
			tmpdata->m_strDescription=*BSTRtoCStr(bstrHelpString);
		WORD fflags=pFuncDesc->wFuncFlags;
		if (!(fflags & (FUNCFLAG_FNONBROWSABLE | FUNCFLAG_FHIDDEN | FUNCFLAG_FRESTRICTED))) {
			SetItemState(newh,TVIS_BOLD,TVIS_BOLD);
		}
		/*if (fflags & (FUNCFLAG_FDEFAULTBIND ))
			SetItemState(newh,TVIS_BOLD,TVIS_BOLD);*/

		
		m_pCurrentTypeInfo->ReleaseFuncDesc(pFuncDesc);
		iIndex++;
	}

	for(iIter = 0; iIter < pTypeAttributes->cVars; iIter++) {
		m_pCurrentTypeInfo->GetVarDesc(iIter, &pVarDesc) ;
		memberID = pVarDesc->memid;
		m_pCurrentTypeInfo->GetDocumentation(memberID, &bstrProperty, &bstrHelpString, NULL, NULL);
		Cstr varval=*BSTRtoCStr(bstrProperty);
		varval+=" : ";
		Cstr gotval("");
		if (pVarDesc->varkind==VAR_CONST)
			GetValue(pVarDesc->lpvarValue,&gotval);
		else
			GetType(&pVarDesc->elemdescVar.tdesc,&gotval);
		varval+=gotval;
		HTREEITEM newh=MyInsertItem(varval.buf,TYPE_CONSTANT,item,iIter);	
		tmpdata->m_strDescription=*BSTRtoCStr(bstrHelpString);

		m_pCurrentTypeInfo->ReleaseVarDesc(pVarDesc);
		iIndex++;
	}

   
	if (pTypeAttributes->typekind==TKIND_COCLASS) {
		HREFTYPE hreftype;  
		LPTYPEINFO ptinfoInterface;  
		LPTYPELIB ptlib = NULL;
		unsigned int nIndex;    
		unsigned short n;   
		TYPEKIND typekind;
		int impltf;
		BSTR bstrName;

		for (n=0; n<pTypeAttributes->cImplTypes; n++)
		{   
			ptinfoInterface = NULL;
			ptlib=NULL;
			if (m_pCurrentTypeInfo->GetRefTypeOfImplType(n, &hreftype)==S_OK)  {
				if (m_pCurrentTypeInfo->GetRefTypeInfo(hreftype, &ptinfoInterface)==S_OK) {   
					if (ptinfoInterface->GetContainingTypeLib(&ptlib, &nIndex)==S_OK) { 
						if (ptlib->GetTypeInfoType(nIndex, &typekind)==S_OK) {    
							if (m_pCurrentTypeInfo->GetImplTypeFlags(n,&impltf)!=S_OK) 
								impltf=0;
							bstrName=NULL;
							ptlib->GetDocumentation(nIndex, &bstrName, NULL, NULL, NULL);
							int typ;        
							if ((impltf & IMPLTYPEFLAG_FSOURCE)==IMPLTYPEFLAG_FSOURCE)
								typ=TYPE_EVENTINTERFACE;
							else if (typekind==TKIND_INTERFACE)
								typ=TYPE_INTERFACE;
							else if (typekind==TKIND_DISPATCH)
								typ=TYPE_DISPINTERFACE;

							newh=MyInsertItem(BSTRtoChar(bstrName),typ,item);	
							if ((pTypeAttributes->wTypeFlags & TYPEFLAG_FCANCREATE)==TYPEFLAG_FCANCREATE ){
								OLECHAR guidstr[128];
								StringFromGUID2(pTypeAttributes->guid,guidstr,128);
								GetProgID(newh,BSTRtoChar(guidstr));
								//if (guidstr!=NULL) ::SysFreeString(guidstr);

							}
						}		
					}
				}
			}
		}
		if (ptinfoInterface!=NULL) ptinfoInterface->Release();
		if (ptlib!=NULL) ptlib->Release();
	}	
	return TRUE;
}

void CTlbTreeCtrl::GetProgID(HTREEITEM item,const char * ID)
{

	HKEY   hKeyTypeLib = NULL;
	HKEY   hKeyClassCLSID = NULL;
	HKEY   hKeyCLSID = NULL;
	HKEY   hKeyProgID = NULL;
	long  dwSize = 128;
	char  ProgID[129];
	HTREEITEM newh;
	if (_RegOpenKey(HKEY_CLASSES_ROOT,"CLSID",&hKeyCLSID)){
		if (_RegOpenKey(hKeyCLSID,ID, &hKeyClassCLSID)) { 
			newh=MyInsertItem(ID,TYPE_CLSID,item);	
			if (_RegOpenKey(hKeyClassCLSID,"ProgID",&hKeyProgID)) {					//Get the version information
				if (RegQueryValue(hKeyProgID,NULL,ProgID,&dwSize)== ERROR_SUCCESS) {
					newh=MyInsertItem(ProgID,TYPE_OBJ,item);
				}
				RegCloseKey(hKeyProgID);
			}
			RegCloseKey(hKeyClassCLSID);
		}			
		RegCloseKey(hKeyTypeLib) ;
	}
}



void  CTlbTreeCtrl::GetValue(VARIANT * v,Cstr * retstr)  
{    
    HRESULT hr; 
    VARIANT vTemp; 
    VariantInit(&vTemp); 
    hr = VariantChangeType(&vTemp, v, 0, VT_BSTR); 
    if (!FAILED(hr)) 
		*retstr=*BSTRtoCStr((V_BSTR(&vTemp)),DONTFREE);  
    VariantClear(&vTemp);
} 


void CTlbTreeCtrl::GetType(TYPEDESC * pTypeDesc,Cstr * rets) 
{   
    VARTYPE vartype= pTypeDesc->vt;   
	BSTR bstrName;
	LPTYPEINFO ptinfoUserDefined;  
	Cstr s("");
	switch (vartype & ~0x7000) 
	{  
		case VT_EMPTY :		s="Empty";break;
		case VT_NULL :		s="Null";break; 
		case VT_I2 :		s="Word";break; 
		case VT_I4 :		s="Int";break; 
		case VT_R4 :		s="Real";break; 
		case VT_R8 :		s="Double";break; 
		case VT_CY :		s="Currency";break; 
		case VT_DATE :		s="Date";break; 
		case VT_BSTR :		s="String";break; 
		case VT_DISPATCH :	s="Dispatch";break; 
		case VT_ERROR :		s="Error";break; 
		case VT_BOOL :		s="Bool";break; 
		case VT_VARIANT :	s="Variant";break; 
		case VT_UNKNOWN :	s="Unknown";break; 

		case VT_I1 :		s="Byte";break; 
		case VT_UI1 :		s="UByte";break; 
		case VT_UI2 :		s="UWord";break; 
		case VT_UI4 :		s="UInt";break; 
		case VT_I8 :		s="Long";break; 
		case VT_UI8 :		s="ULong";break; 
		case VT_INT :		s="Int";break; 
		case VT_UINT :		s="UInt";break; 
		case VT_VOID :		s="Void";break; 
		case VT_HRESULT :	s="Hresult";break; 
		case VT_SAFEARRAY : s="SafeArray";break;   
		case VT_CARRAY :	s="Carray";break; 
		case VT_LPSTR :		s="Lpstr";break; 
		case VT_LPWSTR :	s="Lpwstr";break; 

		case VT_FILETIME :	s="Filetime";break; 
		case VT_BLOB :		s="Blob";break; 
		case VT_STREAM :	s="Stream";break; 
		case VT_STORAGE :	s="Storage";break; 
		case VT_STREAMED_OBJECT : s="Streamed Object";break; 
		case VT_STORED_OBJECT : s="Stored Object";break; 
		case VT_BLOB_OBJECT : s="Blob Object";break; 
		case VT_CF :		s="Cf";break; 
		case VT_CLSID :		s="Clsid";break; 
		case VT_PTR:        GetType(pTypeDesc->lptdesc,&s);
							s=Cstr("*")+s;break; 

		case VT_USERDEFINED:  
			m_pCurrentTypeInfo->GetRefTypeInfo(pTypeDesc->hreftype, &ptinfoUserDefined);  
			ptinfoUserDefined->GetDocumentation(MEMBERID_NIL, &bstrName, NULL, NULL, NULL);
			s=*BSTRtoCStr(bstrName);
			break;
	}     
	Cstr arry("SAFEARRAY(");
	if (vartype & 0x2000)  s=arry+ s +")";   // If SafeArray
    *rets=s;
}    

/////////////////////////////////////////////////////////////////////////////
// CTlbTreeCtrl message handlers


void CTlbTreeCtrl::GetMoreInfo(HTREEITEM item,int typ, bool visit)
{
	if (typ==TYPE_TLB ) {
		GetInformationForTLB(item,visit);
	} else if (typ>=TYPE_ENUM &&  typ<=TYPE_UNION) {
		HTREEITEM parent=GetParentItem(item);
		CTypeLibraryData* ppData=(CTypeLibraryData*)GetItemData(parent);
		if (ppData->type==TYPE_GROUPFOLDER) {
			GetInformationForTLB(GetParentItem(parent),false);
			GetTypeInfoInformation(item,visit);
		}
	}
}

void CTlbTreeCtrl::doOnSelchanged(HTREEITEM itemold, HTREEITEM itemnew) 
{
	if (itemnew==NULL) return;

	CTypeLibraryData* pData=(CTypeLibraryData*)GetItemData(itemnew);
	if (pData!=NULL) {
		int typ=pData->type;
		if (typ>=TYPE_METHOD && typ<=TYPE_CONSTANT) {
			if (!gPlaying) {
				gScintilla->ClearAllTlb();
				gScintilla->AddLine(pData->m_strDescription.buf);
			}
		} else 
			GetMoreInfo(itemnew,typ,false); 
	}
}


void CTlbTreeCtrl::doOnExpand(HTREEITEM hItem) 
{
	CTypeLibraryData* pData=(CTypeLibraryData*)GetItemData(hItem);
	if (pData!=NULL) {
		if (pData->visited) return;
		GetMoreInfo(hItem,pData->type,true);
		pData->visited=true;
	}
}


void CTlbTreeCtrl::doOnDblclk(HTREEITEM hItem) 
{
	CMacroCommunicator * cmnctr;
	if (gTesting)
		cmnctr=gScintilla;
	else
		cmnctr=gMaindialog;
	char * itemtext=GetItText(hItem);
	char * itemtext2=new char[strlen(itemtext)+3];
	sprintf(itemtext2,"\"%s\"",itemtext);
	CTypeLibraryData* pData=(CTypeLibraryData*)GetItemData(hItem);
	if (pData->type==TYPE_HELP) 
		gMaindialog->WindowsOpen(itemtext);
	else if (pData->type==TYPE_CLSID) {
		if (!gMaindialog->RunMacroWithArgs("OnInsertHTMLObject",itemtext2))
			cmnctr->SendText(itemtext);
	} else if (pData->type==TYPE_OBJ) {
		if (!gMaindialog->RunMacroWithArgs("OnInsertScriptObject",itemtext2))
			cmnctr->SendText(itemtext);
	} else if (pData->type>=TYPE_METHOD && pData->type<=TYPE_EVENT) {
		if (!gMaindialog->RunMacroWithArgs("OnInsertFunction",itemtext2))
			cmnctr->SendWithToolTip(itemtext);

	} else if (pData->type==TYPE_TLB) {
		HTREEITEM h=GetParentItem(hItem);
		HTREEITEM hDest;
		if (h==hReg) {
			hDest=hRegFav;
		} else {
			hDest=hReg;
		}
		HTREEITEM newh=InsertItem(GetItText(hItem),TYPE_TLB,TYPE_TLB,hDest,TVI_SORT);	
		SetItemData(newh,(unsigned long)pData);
		pData->visited=false;
		h=GetChildItem(hItem);
		newh=InsertItem(GetItText(h),TYPE_TLBFILE,TYPE_TLBFILE,newh);
		SetItemData(newh,(unsigned long)GetItemData(h));
		HTREEITEM hnext=GetNextSiblingItem(h);					
		DeleteItem(h);
		h=hnext;
		while (h!=NULL) {
			hnext=GetNextSiblingItem(h);
			CleanItemData(h);
			h=hnext;
		}
		DeleteItem(hItem);
		TreeToIni(NULL,hRegFav,SECTION_TLBFAVORITES);		
	} else if (pData->type==TYPE_ENUM || pData->type==TYPE_INTERFACE 
				|| pData->type==TYPE_DISPINTERFACE || pData->type==TYPE_MODULE) {
		HTREEITEM h=GetChildItem(hItem);
		while (h!=NULL) {
			cmnctr->SendText(GetItText(h));
			cmnctr->SendText("\n");
			h=GetNextSiblingItem(h);
		}
	}
	delete [] itemtext2;
}


int CTlbTreeCtrl::GetItemType(HTREEITEM h) 
{
		CTypeLibraryData* pdata=(CTypeLibraryData*)GetItemData(h);
		return pdata->type;
}
