// TreeNode.cpp

#include "stdafx.h"
#include <shlobj.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


extern	int folderIcon;				//closed folder icon
extern	int ofolderIcon;			//open folder icon
extern LPMALLOC gpMalloc;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


TreeNode::TreeNode():
m_pParent(0),
m_pID(0),
m_pFullID(0),
m_pShellFolder(0),
m_explored(false),
m_hassubfolders(false),
m_hItem(NULL),
m_icon(-1),
m_oicon(-1),
m_Folders(0),
m_Files(0),
m_next(NULL)
{	
	SHGetDesktopFolder(&m_pShellFolder);
	if(SHGetSpecialFolderLocation(NULL, CSIDL_DESKTOP, &m_pID)==NOERROR){
		int mySize=GetIDLSize(m_pID);
		m_pFullID = (LPITEMIDLIST)gpMalloc->Alloc(mySize);
		if (m_pFullID){
			CopyMemory(m_pFullID, m_pID, mySize);
			GetFriendlyName(m_pShellFolder,m_pID, SHGDN_INFOLDER,&m_strName);
			GetFriendlyName(m_pShellFolder,m_pID, SHGDN_FORPARSING,&m_strPath);
			GetIcons();
		}
	}
}

TreeNode::TreeNode(TreeNode* pParent, LPITEMIDLIST pID):
m_strName(NULL),
m_strPath(NULL),
m_pParent(pParent),
m_pID(pID),
m_pFullID(0),
m_pShellFolder(0),
m_explored(false),
m_hassubfolders(false),
m_hItem(NULL),
m_icon(-1),
m_oicon(-1),
m_Folders(0),
m_Files(0),
m_next(NULL)
{
	int parentSize = GetIDLSize(m_pParent->m_pFullID) - sizeof(m_pParent->m_pFullID->mkid.cb);
	int mySize = GetIDLSize(m_pID);
	m_pFullID = (LPITEMIDLIST)gpMalloc->Alloc(parentSize + mySize);
	if (m_pFullID){
		CopyMemory(m_pFullID, m_pParent->m_pFullID, parentSize);
		CopyMemory(((LPSTR)m_pFullID) + parentSize, m_pID, mySize);
	}
}

TreeNode::~TreeNode()
{
	gpMalloc->Free(m_pID);
	gpMalloc->Free(m_pFullID);
	if (m_pShellFolder) m_pShellFolder->Release();
	Cleanup();
	if (m_strName!=NULL) delete [] m_strName;
	if (m_strPath!=NULL) delete [] m_strPath;
}

void TreeNode::Cleanup()
{
	TreeNode * p=m_Folders;
	TreeNode * pp;
	while(p!=NULL){pp=p->m_next;delete p;p=pp;}
	p=m_Files;
	while(p!=NULL){pp=p->m_next;delete p;p=pp;}
	m_explored=false;
	m_Files=0;
	m_Folders=0;
}


void TreeNode::GetFriendlyName(LPSHELLFOLDER lpsf,LPITEMIDLIST  lpi,DWORD dwFlags,char ** FriendlyName)
{
   STRRET str;
   char * FName="";
   if (NOERROR==lpsf->GetDisplayNameOf(lpi,dwFlags, &str))
   {
      switch (str.uType)
      {
         case STRRET_WSTR:
			FName=BSTRtoChar(str.pOleStr,false);
			break;
         case STRRET_OFFSET:
             FName=(char *)lpi+str.uOffset;break;
         case STRRET_CSTR:
             FName= (char *)str.cStr;break;
      }
   }
	*FriendlyName=new char [strlen(FName)+1];
	lstrcpy(*FriendlyName,FName);
}

bool TreeNode::CompareIDL(SHITEMID& rId1, SHITEMID& rId2)
{
	UINT nLen = rId1.cb;
	if (nLen != rId2.cb) return false;
	if (nLen == 0) return true;

	nLen -= sizeof(USHORT);
	while (nLen > 0) {
		nLen--;
		if (rId1.abID[nLen] != rId2.abID[nLen]) return false;
	}
	return true;
}


UINT TreeNode::GetIDLSize(LPITEMIDLIST pIDL)
{
	if (!pIDL) return 0;
	char * pIDL0=(char *) pIDL;
	while (pIDL->mkid.cb) 
		pIDL = (LPITEMIDLIST)((char*)pIDL + pIDL->mkid.cb);
	return (UINT)((char *)pIDL-pIDL0)+sizeof(USHORT);
}


void TreeNode::EnumChildren()
{
	LPENUMIDLIST  pEnumIDList = 0;
	ULONG	nReturned, uAttr;
	LPITEMIDLIST  pID = 0;
	TreeNode * lastfolder=0;	//last allocated child folder
	TreeNode * lastfile=0;		//last allocated child file
	Cleanup();
	m_explored=true;

	if (!m_pShellFolder) {
		if (!m_pParent->m_pShellFolder) return;
		if (m_pParent->m_pShellFolder->BindToObject(m_pID, NULL, IID_IShellFolder, (LPVOID*)&m_pShellFolder) != NOERROR)
		return;
	}

	if (m_pShellFolder->EnumObjects(NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS | SHCONTF_INCLUDEHIDDEN, &pEnumIDList) != S_OK)
		return;

	while ((pEnumIDList->Next(1, &pID, &nReturned) != S_FALSE) && (nReturned == 1)) {
		// Select interesting attributes
		uAttr = SFGAO_HASSUBFOLDER | SFGAO_FOLDER | SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR;
		m_pShellFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pID, &uAttr);
		// Only want real Files/Folders or Ancestors : No Control Panel, Printers ...
		if ((uAttr & (SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR))){
			//Create Node
			TreeNode* pChild = new TreeNode(this, pID);

			// Get Name and Path
			GetFriendlyName(m_pShellFolder,pID, SHGDN_INFOLDER ,&pChild->m_strName);
			GetFriendlyName(m_pShellFolder,pID, SHGDN_NORMAL | SHGDN_FORPARSING  ,&pChild->m_strPath);
			// Put in list
			if (uAttr & (SFGAO_HASSUBFOLDER | SFGAO_FOLDER)){
				if (!(uAttr & SFGAO_FILESYSANCESTOR)) {
					pChild->m_icon=folderIcon;
					pChild->m_oicon=ofolderIcon;
				} else
					pChild->GetIcons();
				if (lastfolder==0)
					m_Folders=pChild;			//First subfolder
				else
					lastfolder->m_next=pChild; // link previous to this one
				lastfolder=pChild;
				if (uAttr & (SFGAO_HASSUBFOLDER)) 
					pChild->m_hassubfolders=true;	//It has subfolders
			}
			else{
				pChild->GetIcons(/*true*/);
				if (lastfile==0)
					m_Files=pChild;				// First file
				else
					lastfile->m_next=pChild;	//link previous
				lastfile=pChild;

			}
		}
	} 
	pEnumIDList->Release();
}



TreeNode* TreeNode::FindChild(const char* path)
{
	if (!path) return 0;
	char tmp[MAX_PATH];
	
	//look in folders then in files
	TreeNode * pNode=m_Folders;
	TreeNode * nextroot=m_Files;
	
	while(pNode) {
		lstrcpy(tmp,pNode->m_strPath);
		char * tmplastchar=tmp+lstrlen(tmp)-1;
		if (*tmplastchar=='\\') *tmplastchar='\0';
		if (!stricmp(path, tmp))
			return pNode;
		pNode=pNode->m_next;
		if (pNode==NULL) {
			pNode=nextroot;
			nextroot=NULL;
		}
	}
	return 0;
}

TreeNode* TreeNode::FindRootChild(const char* path)
{
	LPITEMIDLIST pidl;
	int folder=CSIDL_DRIVES;
	if (*path == '\\' && *(path + 1) == '\\')
		folder=CSIDL_NETWORK;
	
	SHGetSpecialFolderLocation(NULL, folder, &pidl);

	TreeNode* pNode;
	TreeNode* retVal=0;
	pNode=m_Folders;
	while(pNode) {
		if (CompareIDL(pNode->m_pID->mkid, pidl->mkid)){
			retVal=pNode;break;
		}
		pNode=pNode->m_next;
	}
	gpMalloc->Free(pidl);
	return retVal;
}


void TreeNode::GetIcons()
{ 
	SHFILEINFO shinfo;
	UINT cbfileinfo=SHGFI_PIDL | SHGFI_SYSICONINDEX | SHGFI_SMALLICON;
	SHGetFileInfo((LPCSTR)m_pFullID,0,&shinfo,sizeof(SHFILEINFO), cbfileinfo);
	m_icon=shinfo.iIcon;
	m_oicon=m_icon;
}
