#if !defined(MACRO_H__INCLUDED_)
#define MACRO_H__INCLUDED_


#define MAX_WORDLEN 100
#define MAX_STACK 20

//macro icons : same as macro flags
#define empty_macro 0
#define bad_macro 1
#define good_macro 2
#define good_called_macro 3
#define good_returning_macro 4
#define good_called_returning_macro 5
#define good_recorded_macro 6
#define calling_macro 256
#define erased_macro 512

class macro;

// Pure virtual Class to handle reception of calls coming from macro dll
class CMacroCommunicator
{
public:
	virtual void ToSendMessage(int typ,const char * arg=NULL,const char * arg2=NULL)=0;	//Send Message to caller
	virtual void ToSendMacroCommand(int cmd,int p1,char * t,char * params)=0;
	virtual int GetInfo(int code,char * params,int arg,int arg2)=0;
	virtual bool FilerxDebug(macro * m,int line)=0;
#ifndef macrodll
	// used from inside Filerx only ( not by macro.dll)
	virtual void SendText(const char * txt)=0;
	virtual void SendWithToolTip(const char * txt)=0;
#endif
};


#ifdef macrodll


class macvar{
public:
	int type;				//combination of TYP_INT,TYP_STRING,TYP_CONST,TYP_EXPR
	char * name;			//variable name
	int iVal;				//integer value
	char * sVal;			//string value
	char * eVal;			//Expression as typed (if TYP_EXPR)
	char * cVal;			//Compiled expression
	int dim;
	int * array;
	macvar * next;			//link to next var
};

//To save vars between calls :
// Only value is saved
class macsavedvar{
public:
	int type;				// TYP_INT or TYP_STRING
	int iVal;				//integer value
	char * sVal;			//string value
	macvar * savedfrom;		//pointer to original variable
	macsavedvar * next;		//link to next
};


struct compilerdata {
	char keyword[MAX_WORDLEN];
	macro * mac;
	int whilestack[MAX_STACK];
	int whilestackptr;
	int ifstack[MAX_STACK];
	int ifstackptr;
	int elsestack[MAX_STACK];
	int elsestackptr;
	int forstack[MAX_STACK];
	int forstackptr;
	int blockstack[5*MAX_STACK];
	int blockstackptr;
	int waitblockbeg;
	int autocloseblock;
	bool gotreturn;
	int gotcall;
	char * params;
	char * macend;
	int ipmac;
	int formac;
	char * linebeg;
	char * lineend;

	int lastnotif;		// last recorded notification message
	int lastwParam;
	int samecount;
	
	void AddBlockToGlobalStack();
	void AddBlockToStack(int f);
	void CloseOneBlock();
	char * GetFor2jtrue(int offset);
	void CompileBreakOrContinue(int type);
	char * PickWord(char * p);
	// Text parsing
	char * FindMacroBeg(char * p);
	char * FindMacroEnd(char * p);
	char * SkipSpaces(char * p);
	char * SkipAllSpaces(char * p);
	char * SkipString(char * p);
	char * FindInstructionEnd(char * p);
	char * FindEol(char * p);

};

class macro {
public:
	char name[MAX_WORDLEN];		//macro name
	int flags;
	int shortcut;
	int breakpoints[10];
	macro * prev;		//link to previous
	macro * next;		//link to next

public:
	macro();
	~macro();
public:
	char inputseparator;
	int len;			//length of code
	int type;			//return type : 0:void;  1:int;  2:string
	int parms;			//number of call parameters
	bool macerror;			// temp err flag : reset for each instruction
	bool gotcompileerror;	//global err flag : true if 1 or more err occured
	//int num;
	compilerdata * cd;	// used for compilation
	char * buff;		// code buffer
	int maxlen;			//total size of code buffer

	//parameter passing /return value
	char parmtyps[10];	//types of call parameters (ex : "IIS")
	macvar * pparms;	//first parameter address
	macvar * retval;	//return value
	macvar * vars;				// variable list
	macsavedvar * savedvars;	// saved variables during call;
	
	//recording
	char * charbuff;			// to accumulate string messages
	
	//compilation
	int errnum;				// error code
	int errpmac;			// instruction pointer
	bool warning;			// true if error is warning
	bool forcewarning;		// true to treat all errors as warnings
	char * errmsg;			// error message
	//int pmac;				// current instruction pointer
	int debuggedline;

	//Main actions
	void Compile(char * macbeg,char ** macend);
	int DecompileInstruction(char ** ip,char ** bend,int lastmsg=0,int crandtab=1);
	char * DecompileTitle(char * buffer);
	char * Decompile();
	void Record(int message,int wParam,int lParam);
	void Play();
	char * DumpCode();

	
	//Add/Remove Instructions to/from code
	void Resize(int needed);
	void AddComment(char * cmnt,int code,char * lend=NULL);
	void AddField(int code);
	void RemoveField(int offset);
	void CompileError(int num,bool iswarning=false);
	void ErrorComment(char * line,char * lend=NULL);
	void AppendIntParm(int value);
	void AppendStringParm(char * value);
	void AppendParm(int parm);
	//Compilation helpers
	char * PickParameters(char *p);
	char * CompileOneInstruction(char * p);
	char * CompileParameters(char * p,char * typs);
	char * CompileExpression(char * p,int typ);
	char * CompileAffect(char * p,macvar * v);
	char * CompileIncrDecr(char * p,macvar * v);
	macvar * DeclareVar(int f,bool toend=false);
	char * CloseBlocks(char * p);
	
	// Variables handling : compilation time
	void DeleteVars();
	void ReaffectString(char ** s,char * value);
	macvar * AddVar(char * name,int typ,int slen=1,bool toend=false);
	macvar * FindVar(char * name);
	void AffectVar(macvar * v,macvar * v2,bool eval=true);
	// Variables handling : execution time
	void DeleteArray(macvar * mv);
	int SaveVars();
	void RestoreVars(int count);
	char * GetValue(char * varname);
	char * GetExpressionValue(char * xpr) ;
	//Expressions
	void Execute(char * s);
	int IEval(char * s);
	char * SEval(char * s);
	int EvalExpression(macvar * mv);
	int CheckEval(macvar * mv,char * dumpbuffer=NULL);

	//Debugging
	bool DoDebug(char * macptr);
	int ip2line(char ** ip,char * limit);

	//General ...
	int EnumFiles(char * spec,bool dirs,macvar * t);

};


#define TYP_INT 1
#define TYP_STRING 2
#define TYP_EXPR 4
#define TYP_CONST 8


struct mdesc {
	int message;
	char * desc;
	char * parms;
};

struct fdesc {
	int message;	//Token code
	char * desc;	//keyword text
	char * parms;	
};

struct cdesc{
	char * desc;	//keyword text
	int value;	
};


//Compiler Errors
#define ERR_MISSINGCPAR 1
#define ERR_TOOMANYMACPARMS 3
#define ERR_BADPARMLIST 4
#define ERR_MISSINGPARM 5
#define ERR_BADPARMTYPE 6
#define ERR_VARNOTDECLARED 7
#define ERR_MISSINGOBRACE 8
#define ERR_SYNTAX 11
#define ERR_MACROISVOID 12
#define ERR_MACRONOTVOID 13
#define ERR_VARMACNOTFOUND 15
#define ERR_OBRACEWANTED 16
#define ERR_MISSINGIF 17
#define ERR_MISSINGCQUOTE 18
#define ERR_BADOP 20
#define ERR_MISSINGOPAR 21
#define ERR_EMPTYEXPR 23
#define ERR_BADVARTYPE 24
#define ERR_TOOCOMPLEX 25
#define ERR_TYPMISMATCH 26
#define ERR_MISSINGOPERAND 27
#define ERR_MISSINGOPORFUNC 28
#define ERR_VARREDEF 29
#define ERR_MACTYPMISMATCH 30
#define ERR_MACPARMTYPMISMATCH 31
#define ERR_ARRAYNOTDECLARED 32
#define ERR_MISSINGMACPARM 33




// Structure of an instruction :
struct macfield{
	unsigned char len;			//length
	unsigned char pnum;			//number of following parameters (may be 0)
	unsigned short  code;		//instruction code
	macvar * mv1;				//parameter 1 points to variable or macro
	macvar * mv2;				//parameter 2 points to variable 
	macvar * mv3;				// .... more parameters might follow
	macvar * mv4;				// .... more parameters might follow
};

#define INSTRLEN0 4 //simple instruction length
#define INSTRLEN1 8	//instruction + 1 parameter
#define PARMLEN 4	//1 parameter

//To set/get info about an instruction

class Instruction{
public:
	Instruction() {};
	Instruction(char *p){field=(macfield *)p;}
public:
	macfield * field;
	macvar ** varenum;
	int Getlen(){return (int)field->len;}
	void Setlen(int l){field->len=(unsigned char)l;}
	int Getpnum(){return (int)field->pnum;}
	void Setpnum(int p){field->pnum=(unsigned char)p;}
	int Getcode(){return (int)field->code;}
	void Setcode(int c){field->code=(unsigned short)c;}
	macvar * Getv1(){return field->mv1;}
	macvar * Getv2(){return field->mv2;}
	macvar * Getv3(){return field->mv3;}
	macvar * Getv4(){return field->mv4;}
	void Setv1(macvar * v){field->mv1=v;}
	void Setv2(macvar * v){field->mv2=v;}
	void Setv3(macvar * v){field->mv3=v;}
	void Setv4(macvar * v){field->mv4=v;}
	macvar * GetFirstVar(){varenum=&(field->mv1);return *varenum; }
	macvar * GetNextVar(){varenum++;return *varenum; }
};

char * Dump1Command(char * iptr,char ** linebuff,char ** rtype);
char * DumpCommands(int kind,int withnumbers,macro * mac);
int typfromletter(char * c);

__declspec( dllexport ) void SetCommunicator(CMacroCommunicator * maccom);
__declspec( dllexport ) bool LoadIface(char * ifacename);
__declspec( dllexport ) macro * CreateMacro();
__declspec( dllexport ) void DeleteMacro(macro * m);
__declspec( dllexport ) int GetCommandCode(char * command);
__declspec( dllexport ) char * GetMacroKeyWords(int flag);
__declspec( dllexport ) char * GetToolTip(char * fname,macro * macs);
__declspec( dllexport ) void Compile(macro * m,char * macbeg,char ** macend);
__declspec( dllexport ) void Play(macro * m);
__declspec( dllexport ) void Debug(macro * m);
__declspec( dllexport ) char * GetValueOf(macro * m,char * varname);
__declspec( dllexport ) char * GetValueOfExpr(macro * m,char * expr);
__declspec( dllexport ) void Record(macro * m,int message,int wParam,int lParam);			
__declspec( dllexport ) char * Decompile(macro * m,macro * ms);		
__declspec( dllexport ) char * Dumpcode(macro * m);
__declspec( dllexport ) void PrepareCopyData(bool p);

#else //#ifndef macrodll

// What the dll user sees from outside :

class macro {
public:
	char name[MAX_WORDLEN];		//macro name
	int flags;					//compilation/type .
	int shortcut;
	int breakpoints[10];
	macro * prev;				//link to previous
	macro * next;				//link to next
};

__declspec( dllimport ) void SetCommunicator(CMacroCommunicator * maccom);
__declspec( dllimport ) bool LoadIface(char * ifacename);
__declspec( dllimport ) macro * CreateMacro();
__declspec( dllimport ) void DeleteMacro(macro * m);
__declspec( dllimport ) int GetCommandCode(char * command);
__declspec( dllimport ) char * GetMacroKeyWords(int flag);
__declspec( dllimport ) char * GetToolTip(char * fname, macro * macs);
__declspec( dllimport ) bool Compile(macro * m,macro * ms,char * beg,char ** end);
__declspec( dllimport ) void Play(macro * m);
__declspec( dllimport ) void Debug(macro * m);
__declspec( dllimport ) char * GetValueOf(macro * m,char * varname);
__declspec( dllimport ) char * GetValueOfExpr(macro * m,char * expr);
__declspec( dllimport ) void Record(macro * m,int command,int wParam,int lParam);
__declspec( dllimport ) char * Decompile(macro * m,macro * ms);
__declspec( dllimport ) char * Dumpcode(macro * m);
__declspec( dllimport ) void PrepareCopyData(bool p);

#endif



#endif //MACRO_H__INCLUDED_