#include "stdafx.h"
#include "..\common\defines.h"
#include "..\common\macro.h"
#include "maccore.h"
#include <time.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern macro * macros;
extern int xcommands;
extern mdesc * xcommandtable;
extern int SearchKeywordByName(char *, char **);

bool scripting;
/////////////////////////////////////////////////////////////////////////////
//	COMPILING FUNCTIONS


// resize macro code buffer by increments of 200

void macro::Resize(int needed)
{
	bool mustresize=false;
	while (len + needed>= maxlen) {
		maxlen += 200;
		mustresize=true;
	}
	if (mustresize){
		char *newBuffer = new char[maxlen];
		memcpy(newBuffer, buff, len);
		delete [](buff);
		buff = newBuffer;
	}
}


void macro::AddComment(char * cmnt,int code,char * lend/*=NULL*/)
{
	char * p=cmnt;
	if (*p=='/' && *(p+1)=='/'){
		// This is a triple slashed comment : error comment added by compiler
		// Don't compile it ...
		if (*(p+2)=='/') return;
		p+=2;
	}

	char * commbuff;
	if (lend!=NULL) {
		// remove trailing spaces
		char * llend=lend;
		while (*(llend-1)==' ' || *(llend-1)=='\t') llend--;
		commbuff=new char[llend-p+1];
		strncpy(commbuff,p,llend-p);
		*(commbuff+(llend-p))='\0';
	} else
		commbuff=p;

// A comment is a constant string expression

	AddField(code);
	AddVar(NULL,TYP_STRING|TYP_CONST|TYP_EXPR,strlen(commbuff)+1);
	strcpy(vars->sVal,commbuff);
	vars->eVal=new char[strlen(commbuff)+3];
	sprintf(vars->eVal,"\"%s\"",commbuff);
	AppendParm((int)vars);

	if (lend!=NULL)
		delete [] commbuff;

}

// In case of compilation error, last instruction(s) may be bad
// and will be removed

void macro::RemoveField(int offset)
{
	len=offset;
}

// All compilation errors should generate a call to CompileError
// iswarning is always false ... ???
// When a compilation error occurs :
// - CompileError is called at once to prepare the error message and
//		set macerror flag
// - when compiler's execution reaches the end of CompileOneInstruction
//		error messages are compiled

void macro::CompileError(int num,bool iswarning)
{
	
	int lp;
	int ll;
	char * s;
	errnum=num;
	warning=iswarning;
	if (cd->formac!=NULL) {
		errpmac=cd->formac;
	} else
		errpmac=cd->ipmac;
	bool havs=false;
	switch (errnum){
		case ERR_MISSINGCPAR:s="/Missing ')'";break;
		//case 2:s="/Empty parameter";break;
		case ERR_TOOMANYMACPARMS:s="/Too many macro parameters";break;
		case ERR_BADPARMLIST:s="/Bad Parameter list";break;
		case ERR_MISSINGPARM:s="/Missing parameter(s)";break;
		case ERR_BADPARMTYPE:s="/Bad Parameter type : \'%s\'";break;
		case ERR_VARNOTDECLARED:s="/Variable \'%s\' not declared";havs=true;break;
		case ERR_MISSINGOBRACE:s="/\'}\' without \'{\'";break;
		//case 9:s="/Bad \'for\' syntax";break;
		//case 10:s="/Increment Decrement only numbers";break;
		case ERR_SYNTAX:s="/Syntax Error";break;
		case ERR_MACROISVOID:s="/Void macro can\'t return a value";break;
		case ERR_MACRONOTVOID:s="/This macro must return a value";break;
		//case 14:s="/Missing condition";break;
		case ERR_VARMACNOTFOUND:s="/Variable or macro \'%s\' not found";havs=true;break;
		case ERR_OBRACEWANTED:s="/Missing \'{\'";break;
		case ERR_MISSINGIF:s="/\'else\' without \'if\'";break;
		case ERR_MISSINGCQUOTE:s="/Missing Ending Quote";break;
		case ERR_BADOP:s="/Bad Operator";break;
		case ERR_MISSINGOPAR:s="/Missing \'(\'";break;
		//case 22:s="/Missing \')\'";break;
		case ERR_EMPTYEXPR:s="/Empty Expression";break;
		case ERR_BADVARTYPE:s="/Bad variable type";break;
		case ERR_TOOCOMPLEX:s="/Expression too complex";break;
		case ERR_TYPMISMATCH:s="/Type mismatch";break;
		case ERR_MISSINGOPERAND:s="/Missing Operand in expression";break;
		case ERR_MISSINGOPORFUNC:s="/Missing Operator/Function in expression";break;
		case ERR_VARREDEF:s="/Redefinition of \'%s\' with different type";havs=true;break;
		case ERR_MACTYPMISMATCH:s="/macro type mismatch";break;
		case ERR_MACPARMTYPMISMATCH:s="/macro parameter type mismatch";break;
		case ERR_ARRAYNOTDECLARED:s="/Array \'%s\' not declared";havs=true;break;
		case ERR_MISSINGMACPARM:s="/missing macro parameter(s)";break;
		default :s="/Error ...";break; //should not happen ...
	}

	if (havs)
		lp=strlen(cd->keyword);
	else 
		lp=0;

	ll=strlen(s)+lp+1;
	if (errmsg!=NULL) delete[]errmsg;
		errmsg=new char [ll];
	
	if (!havs)
		strcpy(errmsg,s);
	else
		sprintf(errmsg,s,cd->keyword);
	macerror=true;
	gotcompileerror=true;
}

// called by CompileOneInstruction to compile errors
// - Add a MAC_WARNING instruction to describe error
// - Add a MAC_ERROR instruction containing the bad line

void macro::ErrorComment(char * line,char * lend/*=NULL*/){

	if (warning){ 
		AddComment(errmsg,MAC_WARNING);			//insert
	}else{
		if (errpmac!=-1)
			RemoveField(errpmac);
		AddComment(errmsg,MAC_WARNING);
		AddComment(line,MAC_ERROR,lend);
	}	
}

// Append a new instruction to code buffer

void macro::AddField(int code)
{
	int needed=INSTRLEN0;
	cd->ipmac=len;
	Resize(needed);
	macfield *mf=(macfield *) (buff+cd->ipmac);
	mf->len=needed;
	mf->code=code;
	mf->pnum=0;	//parms
	len+=needed;	//total macro len
	if (code==MAC_JUMP)
		AppendParm(0);
}

// Calculate the length of an expression when it will be slashed

static int CalcSlashLen(const char * src){
	char * s=(char *)src;
	int slashedlen=0;
	while(*s){
		switch (*s){
			case '\a':
			case '\b':
			case '\f':
			case '\n':
			case '\r':
			case '\t':
			case '\v':
			case '\\':
				slashedlen+=2;
				break;
			default:
				slashedlen++;
				break;
		}
		s++;
	}
	return slashedlen;
}

// Used to slash expressions coming from Record and scriptlines

static int SlashCopy(char * dest,const char * src){
	char * s=(char *)src;
	char * d=dest;
	while(*s){
		switch (*s){
			case '\a':
				*d++='\\';*d++='a';s++;
				break;
			case '\b':
				*d++='\\';*d++='b';s++;
				break;
			case '\f':
				*d++='\\';*d++='f';s++;
				break;
			case '\n':
				*d++='\\';*d++='n';s++;
				break;
			case '\r':
				*d++='\\';*d++='r';s++;
				break;
			case '\t':
				*d++='\\';*d++='t';s++;
				break;
			case '\v':
				*d++='\\';*d++='v';s++;
				break;
			case '\\':
				*d++='\\';*d++='\\';s++;
				break;
			default:
				*d++=*s++;
				break;
			s++;
		}
	}
	*d='\0';
	return(d-dest);
}

// Append a string parameter to current instruction
// called by Record and CompileOneInstruction when compiling a script line


void macro::AppendStringParm(char * value)
{
	AddVar(NULL,TYP_STRING|TYP_CONST|TYP_EXPR,strlen(value)+1);
	strcpy(vars->sVal,value);
	int stringlen=CalcSlashLen(value);
	vars->eVal=new char[stringlen+3];
	*vars->eVal='\"';
	SlashCopy(vars->eVal+1,value);
	char * eValend=vars->eVal+strlen(vars->eVal);
	*eValend='\"';
	*(eValend+1)='\0';
	AppendParm((int)vars);
}

// Append an integer parameter to current instruction
// called by Record

void macro::AppendIntParm(int value)
{	
	AddVar(NULL,TYP_INT|TYP_CONST|TYP_EXPR);
	vars->iVal=value;
	char tbuf[15];
	sprintf(tbuf,"%i",value);
	vars->eVal=new char[strlen(tbuf)+1];
	strcpy(vars->eVal,tbuf);
	AppendParm((int)vars);
}

// Append a parameter to current instruction
// called by CompileAffect, CompileIncrDecr, CompileOneInstruction, Record
// and CompileExpression

void macro::AppendParm(int parm)
{	
	int needed=PARMLEN;
	Resize(needed);
	macfield *mf=(macfield *) (buff+cd->ipmac);
	mf->len+=needed;
	mf->pnum+=1;	//parms
	int * mv=(int *) (buff+len);
	*mv=parm;
	len+=needed;
}

// Compile instruction parameters
// called by CompileOneInstruction

char * macro::CompileParameters(char * pp,char * typs)
{	
	char *typptr=typs;
	int ttyp=TYP_INT;
	char * p=pp;
	char typeletter;

	while (*typptr!='\0' && !macerror){
		if (p<cd->lineend && *p!=')') {
			p++; //skip '(' or ','
			p=cd->SkipSpaces(p);
		}
		while  (*typptr=='0') typptr++;
		if (*typptr=='\0') break;
		ttyp=typfromletter(typptr);
		typeletter=*typptr;
		typptr++;
		if (*p==',' || *p==')' || p==cd->lineend) {
			if (typeletter=='I')
				CompileExpression("0",ttyp);
			else if (typeletter=='S')
				CompileExpression("\"\"",ttyp);
		} else {
			 p=CompileExpression(p,ttyp);//points to ',' or ')'
		}
	}
	return cd->lineend;
}

// Compile variable affectation
// called by CompileOneInstruction

char * macro::CompileAffect(char * p,macvar * v){
	if (v!=NULL){
		if (v->type==TYP_INT)
			AddField(MAC_AFFECTINT);
		else
			AddField(MAC_AFFECTSTRING);
		AppendParm((int)v);//var
		p++;
		p=CompileExpression(p,v->type);
	}else //toaffect=NULL
		CompileError(ERR_VARNOTDECLARED);
	return p;
}

// Compile variable xxxx++ and xxxx--
// called by CompileOneInstruction


char * macro::CompileIncrDecr(char * p,macvar * v){
	if (v!=NULL){
		char incdec=*p;
		p=cd->SkipSpaces(p);
		p++;
		if (*p==incdec){
			if (v->type & TYP_INT){
				if (incdec=='+')
					AddField(MAC_INCR);
				else
					AddField(MAC_DECR);
				AppendParm((int)v);
				p++;
			}else //TYP_STRING
				CompileError(ERR_TYPMISMATCH);
		}else // + ??? or - ???
			CompileError(ERR_SYNTAX);
	}else //toaffect=NULL
		CompileError(ERR_VARNOTDECLARED);
	return p;
}


// compile variable declaration
// used by CompileOneInstruction and PickParameters
// toend is true when called by PickParameters
// then variables are added to the end of the list

macvar * macro::DeclareVar(int f,bool toend/*=false*/){
	int vtyp;
	if (f==MAC_INT)
		vtyp=TYP_INT;
	else 
		vtyp=TYP_STRING;

	macvar * fv=FindVar(cd->keyword);
	if (fv!=NULL){
		if (fv->type != vtyp)
			CompileError(ERR_VARREDEF);
	}else
		fv=AddVar(cd->keyword,vtyp,1,toend);
	return fv;
}

// Find a called macro by name
// used by CompileOneInstuction or CompileExpression  when compiling macro calls

macro * FindMacroByName(char * mname,macro * mm){
	if (!lstrcmp(mm->name,mname))
		return mm;
	macro * m=macros;
	while (m!=NULL){
		if (!lstrcmp(m->name,mname)) return m;
		m=m->next;
	}
	return NULL;
}


// Compiles closing braces
// May add closing block instructions, too
// if macend is reached and unclosed blocks exist
// called by main Compile routine


char * macro::CloseBlocks(char * p)
{
	p--;

	while (1){
		if (p<cd->macend)
			p=cd->SkipAllSpaces(p+1);
		if (*p=='#') p++;

		bool endblock=(*p=='}') || (p>=cd->macend && cd->blockstackptr>=0);
		if (!endblock)
			break;

/*		AddField(MAC_BLOCKEND);
		if (cd->blockstackptr<0){
			errpmac=-1;
			CompileError(ERR_MISSINGOBRACE);
			ErrorComment("}");
			macerror=false;
		}else{
			cd->CloseOneBlock();
		}*/
		if (cd->blockstackptr>=0){
			AddField(MAC_BLOCKEND);
			cd->CloseOneBlock();
		} else if (gotcompileerror)
			AddField(MAC_BLOCKEND);
	}
	return p;
}

// Called in a loop by Compile
// and recursively by itself to compile "for" instruction

char * macro::CompileOneInstruction(char * p)
{
	macro * foundm;			//Called  macro
	macvar * newvar;
	int found;

	char * psave=p;

	p=cd->SkipSpaces(p);
	//copy keyword and skip trailing spaces
	p=cd->PickWord(p);
	//Find keyword in list
	found=SearchKeywordByName(cd->keyword,&cd->params);
	
	if (scripting && found!=MAC_SCRIPTEND) {
		AddField(MAC_SCRIPTLINE);
		p=psave;
		while (*psave!='\n' && *psave!='\0') {
			if (*psave=='\r') *psave=' ';
			psave++;
		}
		if (*psave!='\0')	{
			psave++;
			cd->lineend=psave-1;
		}
		char psavechar=*psave;
		*psave='\0';		
		AppendStringParm(p);
		*psave=psavechar;
	//if not found ...
	} else if (found==MAC_COMMENT){
		//may be variable
		macvar * toaffect=FindVar(cd->keyword);
		if (*p=='='){
			p=CompileAffect(p,toaffect);					//var=expr
		}else if (*p=='['){
			if (toaffect!=NULL){
				int tafftyp=TYP_INT;
				int field;
				if (toaffect->dim==0)
					field=MAC_AFFECTCHAR;					//string[index];
				else {
					if ((toaffect->type & TYP_INT)==TYP_INT)
						field=MAC_AFFECTINTARRAY;			//intarray[index]
					else {
						field=MAC_AFFECTSTRINGARRAY;		//stringarray[index]
						tafftyp=TYP_STRING;
					}
				}
				AddField(field);
				AppendParm((int)toaffect);
				p=CompileExpression(p+1,TYP_INT);
				if (*p!=']')
					CompileError(ERR_SYNTAX);
				else {
					p=cd->SkipSpaces(p+1);
					if (*p=='='){
						p=CompileExpression(p+1,tafftyp);	//xxxxxxx=expr
					} else
						CompileError(ERR_SYNTAX);
				}
			} else
				CompileError(ERR_VARNOTDECLARED);

		}else if (*p=='+' || *p=='-'){
			p=CompileIncrDecr(p,toaffect);
		}else //no = ... no ++/--
			CompileError(ERR_SYNTAX);
	}else if (found !=MAC_EMPTY){
		// keyword is known
			AddField(found);
		switch (found){
			case MAC_INT:		// declaration
			case MAC_STRING:
				p=cd->PickWord(p);
				newvar=DeclareVar(found);
				AppendParm((int)newvar);
				if (*p=='='){
					Instruction curi(buff+cd->ipmac);
					if (newvar==vars)					//newly created
						curi.Setcode(found+1);			//becomes MAC_???AFFECT
					else
						curi.Setcode(found+2);			//becomes MAC_AFFECT???
					p=CompileExpression(p+1,newvar->type);
				} else if (*p=='['){
					Instruction curi(buff+cd->ipmac);
					if (found==MAC_INT)
						curi.Setcode(MAC_INTARRAY);		//becomes MAC_???ARRAY
					else
						curi.Setcode(MAC_STRINGARRAY);	//becomes MAC_???ARRAY
					p=CompileExpression(p+1,TYP_INT);
					if (*p!=']')
						CompileError(ERR_SYNTAX);
					else {
						newvar->dim=-1;					//will be allocated later
					}
				}
				break;
			case MAC_SCRIPTBEG:
				p=cd->PickWord(p);
				newvar=DeclareVar(MAC_STRING);
				AppendParm((int)newvar);				
				scripting=true;
				while (*p!='\r' && *p!='\n') p++;
				while ((*p=='\r' || *p=='\n') && *p!='\0') p++;
				cd->lineend=p-1;
				break;
			case MAC_SCRIPTEND:
				scripting=false;
				break;
			case MAC_FOR:		//for
				cd->formac=cd->ipmac;
				cd->AddBlockToStack(MAC_FOR);
				if (*p=='(') p++;
				p=CompileOneInstruction(p);
				if (!macerror) {
					p=cd->SkipSpaces(p);
					if (*p==';') p++;
					AddField(MAC_FOR2);
					p=CompileExpression(p,TYP_INT);
					if (!macerror) {
						AddField(MAC_JUMP);
						AddField(MAC_JUMP);
						p=cd->SkipSpaces(p);
						if (*p==';') p++;
						p=CompileOneInstruction(p);
						if (!macerror) {
							p=cd->SkipSpaces(p);
							if (*p==')') p++;
							AddField(MAC_JUMP);
							cd->formac=NULL;
						}
					}
				}
				if (macerror) {
					cd->ipmac=cd->formac;
					cd->lineend=cd->FindEol(p);
					CompileError(errnum);
					cd->forstackptr--;
					cd->waitblockbeg=0;
				}
				cd->formac=NULL;
				break;
			case MAC_RETURN:			
				if (*p=='('){
					if (type==0)
						CompileError(ERR_MACROISVOID);
					else{
						p=CompileExpression(p+1,type);
						if (*p==')') p++;
						cd->gotreturn=true;
					}
				}else
					if (type!=0)
						CompileError(ERR_MACRONOTVOID);
				break;
			case MAC_BREAK:
			case MAC_CONTINUE:
				cd->CompileBreakOrContinue(found);
				break;
			case MAC_WHILE:
			case MAC_IF:
				if (*p=='('){ 
					p=cd->SkipSpaces(p+1);
				}
				if (*p==')' || *p=='{' || p==cd->lineend)
					CompileExpression("1",TYP_INT);
				else
					p=CompileExpression(p,TYP_INT);//points to ',' or ')'
				if (*p==')') p++;
				if (!macerror) {
					cd->AddBlockToStack(found);
					AddField(MAC_JUMP);
				}
				break;
			case MAC_ELSE:
				p=cd->SkipAllSpaces(p);
				if (cd->blockstack[cd->blockstackptr+1]!=MAC_IF)
					CompileError(ERR_MISSINGIF);
				else{
					char * ifpos=buff+cd->ifstack[cd->ifstackptr+1];
					char * elsepos=ifpos+2*INSTRLEN1-PARMLEN;
					*((int *) elsepos)=len+INSTRLEN1;						//re-patch if
					if (*p=='i' && *(p+1)=='f'){
						cd->lineend=p-1;
					}else{
						cd->AddBlockToStack(found);
					}
					AddField(MAC_JUMP);
				}
				break;
			case MAC_CALL :		//macro call
				cd->gotcall=calling_macro;
				p=cd->PickWord(p);
				foundm=FindMacroByName(cd->keyword,this);
				if (foundm!=NULL){
					AppendParm((int)foundm);
					p=CompileParameters(p,foundm->parmtyps);
				}else
					CompileError(ERR_VARMACNOTFOUND);
				break;
			default:
				//existing keyword
				AppendParm((int) cd->params);
				p=CompileParameters(p,cd->params);
		}//switch
	} else {
		CompileError(ERR_SYNTAX);
	}
	if (macerror)
		ErrorComment(cd->linebeg,cd->lineend);
	return p;
}

// Compile Macro declaration line parameters

char * macro::PickParameters(char *p)
{
	int found;
	//call parameters
	parms=0;
	while (*p=='(' || *p==','){
		p++;
		p=cd->SkipSpaces(p);
		if (*p!=')' && isalpha(*p)){

		}else
			CompileError(ERR_BADPARMLIST);
		//copy keyword ...
		p=cd->PickWord(p);
		//Find keyword in list
		found=SearchKeywordByName(cd->keyword,&cd->params);
		if (found==MAC_INT || found==MAC_STRING){
			p=cd->PickWord(p);
			macvar * fv=DeclareVar(found,true);//AddToEnd
			parmtyps[parms++]=*(cd->params);
			if (*p=='['){
				p=cd->SkipSpaces(p+1);
				if (*p==']'){
					p=cd->SkipSpaces(p+1);
					fv->dim=-1;
				}
			} else
				fv->type|=(TYP_EXPR|TYP_CONST);
		}else
			CompileError(ERR_BADPARMTYPE);
	}
	parmtyps[parms]='\0';
	pparms=vars;

	//return value
	if (type!=0){
		retval=AddVar(NULL,type);
	}else
		retval=NULL;
	return p;
}

// Main compilation routine
// returns macro text end pointer (buffer end or next macro start in buffer)

void macro::Compile(char * macbeg,char ** setmacend){
	char * curpmac;

	DeleteVars();
	len=0;		//compiled macro len=0
	*name='\0';
	flags=empty_macro;	//empty
	curpmac=macbeg;
	macerror=false;
	gotcompileerror=false;
	scripting=false;

	cd=new compilerdata;
	cd->mac=this;
	cd->whilestackptr=-1;
	cd->ifstackptr=-1;
	cd->elsestackptr=-1;
	cd->forstackptr=-1;
	cd->blockstackptr=-1;
	cd->macend=macbeg+strlen(macbeg); //temp
	cd->gotreturn=false;
	cd->formac=NULL;
	cd->gotcall=0;
	cd->autocloseblock=0;
	
	curpmac=cd->FindMacroBeg(curpmac);
	if (curpmac>=cd->macend){
		*setmacend=cd->macend;
		delete cd;
		return;
	}
	*setmacend=cd->FindMacroEnd(curpmac);
	curpmac=PickParameters(curpmac);
	cd->lineend=curpmac;


	while (curpmac<cd->macend){
		curpmac=cd->lineend+1;	
		macerror=false;	
		cd->waitblockbeg=0;
		if (!scripting) {
			curpmac=CloseBlocks(curpmac);
		}
		cd->ipmac=-1;			//no instruction inserted

		if (curpmac>=cd->macend) break;
		cd->linebeg=curpmac;
		//find eol
		if (*curpmac=='/' && *(curpmac+1)=='/'){
			cd->lineend=cd->FindEol(curpmac);
			AddComment(cd->linebeg,MAC_COMMENT,cd->lineend);
		} else {
			cd->lineend=cd->FindInstructionEnd(curpmac);

			if (*curpmac=='#') 
				{curpmac++;cd->linebeg++;}
			
			curpmac=CompileOneInstruction(curpmac);

			if (cd->waitblockbeg!=0 && !scripting){
				curpmac=cd->SkipAllSpaces(curpmac);
				cd->lineend=curpmac;
				if (*curpmac!='{') {
					cd->lineend--; // no error : continue
					cd->autocloseblock++;
				}
				cd->AddBlockToGlobalStack();
				AddField(MAC_BLOCKBEG);
			} else
				while (cd->autocloseblock>0) {
					AddField(MAC_BLOCKEND);
					cd->CloseOneBlock();
					cd->autocloseblock--;
				}
		}		
	}//while ...

	if (gotcompileerror)
		flags=bad_macro;
	else if (len!=0) {
		flags=good_macro | cd->gotcall;
		if (parms!=0) flags++;
		if (type!=0) flags+=2;
	}

	if (type!=0 && !cd->gotreturn){
		cd->ipmac=-1;
		AddComment("/This macro must return a value",MAC_WARNING);
		retval->iVal=0;
		retval->sVal=new char[1];
		*(retval->sVal)='\0';
	}
	delete cd;
}

///////////////////////////////////////////////////////////////////////////////////////

// Compiles macro name and return type
// called by Compile

char * compilerdata::FindMacroBeg(char * p)
{
	// Find Macro first line :  "Macro xxxx"
	char * temppmac;
	int found;
	while (1) {
		p=SkipAllSpaces(p);
		if (p>=macend)	//empty macro 
			return p;
		p=PickWord(p);
		if (!lstrcmp(keyword,"Macro")) {
			p=SkipSpaces(p);
			temppmac=p;
			p=PickWord(p);
			found=SearchKeywordByName(keyword,&params);
			if (found==MAC_INT)
				mac->type=1;
			else if (found==MAC_STRING)
				mac->type=2;
			else { 
				mac->type=0;
				if (found!=MAC_VOID) // type omitted ...
					p=temppmac;
			}
			break;
		}
		p=FindEol(p);	
	}
	p=PickWord(p);
	if (*keyword=='\0') 
		strcpy(mac->name,"unnamed_macro");
	else
		strcpy(mac->name,keyword);
	p=SkipSpaces(p);
	return p;
}	

//Find macro end : "Macro xxxx" or end of buffer
// called by Compile

char * compilerdata::FindMacroEnd(char * p)
{
	char *srchmacend=p;
	char * nextline;
	while (1) {
		srchmacend=FindEol(srchmacend);	
		srchmacend=SkipAllSpaces(srchmacend);
		nextline=srchmacend;
		if (srchmacend>=macend) break;
		srchmacend=PickWord(srchmacend);
		if (!lstrcmp(keyword,"Macro")) {
			break;
		}
	}
	macend=nextline;
	return nextline;
}

// Closes one for/while/if/else block
// called by Compile and CloseBlocks
// adjusts Jump instructions around block start

void compilerdata::CloseOneBlock()
{
	int blk=blockstack[blockstackptr];
	char * buff=mac->buff;
	int len=mac->len;
	blockstackptr--;
	if (blk==MAC_WHILE){
		mac->AddField(MAC_JUMP);
		buff=mac->buff;len=mac->len;
		char * wpos=buff+(whilestack[whilestackptr]+2*INSTRLEN1-PARMLEN); //while jump parm				
		*((int *) wpos)=len;
		*((int *) (buff+len-PARMLEN))=whilestack[whilestackptr];
		whilestackptr--;
	}else if (blk==MAC_IF){
		char * ifpos=buff+(ifstack[ifstackptr]);				
		char * ifjpos=ifpos+2*INSTRLEN1-PARMLEN; //if jump parm				
		*((int *) ifjpos)=len;
		ifstackptr--;
		char * elsepos=ifpos-INSTRLEN0-INSTRLEN1; // else+jump
		char * elsejpos=ifpos-PARMLEN; // else jump pos 
		Instruction ielse(elsepos);
		if (ielse.Getcode()==MAC_ELSE)	//was "else if"
			*((int *) elsejpos)=len;			//patch else jump 
	}else if (blk==MAC_ELSE){
		char * elsepos=buff+elsestack[elsestackptr];
		char * elsejpos=elsepos+INSTRLEN0+INSTRLEN1-PARMLEN; //else jump parm				
		*((int *) elsejpos)=len;
		elsestackptr--;
	}else if (blk==MAC_FOR){
		mac->AddField(MAC_JUMP);
		buff=mac->buff;len=mac->len;
		char * for2jtrue=GetFor2jtrue(forstack[forstackptr]);
		forstackptr--;
		char * for2jfals=for2jtrue+INSTRLEN1;
		char * for3=for2jfals+PARMLEN;
		Instruction ifor3(for3);
		char * for3j=for3+ifor3.Getlen()+INSTRLEN1-PARMLEN;
		*((int *) (buff+len-PARMLEN))=for3-buff;	//last jump
		*((int *) (for2jtrue))=for3j-buff+INSTRLEN1;	//true jump
		*((int *) (for2jfals))=len;			//false jump
		*((int *) (for3j))=for2jtrue-buff-INSTRLEN0-INSTRLEN1;//test			
	}

}

// When '{' is (or should be found), write to global block stack
// called by Compile

void compilerdata::AddBlockToGlobalStack()
{
	blockstackptr++;
	blockstack[blockstackptr]=waitblockbeg;
}

// When compiling a for/if/else/while
// add ip to corresponding stack
// set waitblockbeg to a non zero value
// to indicate a '{' is wanted somewhere ...
// called by CompileOneInstruction

void compilerdata::AddBlockToStack(int f)
{
	if (f==MAC_WHILE){
		whilestackptr++;
		whilestack[whilestackptr]=ipmac;
	}else if (f==MAC_IF){
		ifstackptr++;
		ifstack[ifstackptr]=ipmac;			
	}else if (f==MAC_FOR){
		forstackptr++;
		forstack[forstackptr]=ipmac; //for instruction			
	}else if (f==MAC_ELSE){
		elsestackptr++;
		elsestack[elsestackptr]=ipmac;
	}
	waitblockbeg=f;
}

// calculate jump offsets ...
// called by CloseOneBlock and CompileBreakOrContinue

char * compilerdata::GetFor2jtrue(int offset){
	char * forpos=mac->buff+offset; //for
	char * for1=forpos+INSTRLEN0;//skip for
	Instruction ifor1(for1);
	char * for2=for1+(ifor1.Getlen());//first instruction
	return(for2+2*INSTRLEN1-PARMLEN);
}

// Compile break and continue ...
// called by CompileOneInstruction

void compilerdata::CompileBreakOrContinue(int type)
{
	int a1=0;
	int a2=INSTRLEN1;
	if (type==MAC_BREAK){
		a1=INSTRLEN1;
		a2=0;
	}
	int blkptr=blockstackptr;
	int blk;
	while (blkptr>=0){
		blk=blockstack[blkptr];
		if (blk==MAC_WHILE || blk==MAC_FOR) break;
		blkptr--;
	}
	if (blkptr>=0){
		mac->AddField(MAC_JUMP);
		if (blk==MAC_WHILE)
			*((int *) (mac->buff+mac->len-PARMLEN))=whilestack[whilestackptr]+a1;
		else if (blk==MAC_FOR)
			*((int *) (mac->buff+mac->len-PARMLEN))=(GetFor2jtrue(forstack[forstackptr])-mac->buff)+PARMLEN+a2;
	}
}

//Parsing helpers
//On entry p points to current parsing position
//Returns updated parsing pointer

char * compilerdata::PickWord(char * p){
	char * keywordcopy=keyword;
	int wordlen=0;
	while((isalnum(*p)|| *p=='_')) {
		*keywordcopy++=*p++;
		wordlen++;
		if (wordlen==MAX_WORDLEN-1)
			break;
	}
	*keywordcopy='\0';
	//skip spaces
	return SkipSpaces(p);
}


char * compilerdata::FindInstructionEnd(char * p){
	while (*p!='\r' && *p!='\n' && *p!=';' && *p!='\0') {
		if (*p=='\"' || *p=='\'') 
			p=SkipString(p);
		else
			p++;
	}
	return p;
}

char * compilerdata::FindEol(char * p){
	while (*p!='\r' && *p!='\n' && *p!='\0')
		p++; 
	return p;
}

char * compilerdata::SkipSpaces(char * p)
{
	while(*p==' ' || *p=='\t') p++;
	return p;
}
char * compilerdata::SkipAllSpaces(char * p)
{
	while(*p==' ' || *p=='\t' || *p=='\r' || *p=='\n') p++;
	return p;
}

char * compilerdata::SkipString(char * p)
{
	char limitchar=*p;
	char * e=p;
	e++;
	while(1){
		if (*e=='\\')
			e++;
		else if (*e==limitchar || *e=='\0' || *e=='\r' || *e=='\n'){
			if (*e==limitchar) e++;
			return e;
		}
		e++;
	}
}

