// macro.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "..\common\defines.h"
#include "..\common\macro.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


int calling;
bool stopped;
bool debugging;
bool preparecopydata;
//bool gotcompileerror;
macro * macros;
macro * dynmacro;

mdesc * xcommandtable;
fdesc  *xfunctiontable;
cdesc * constanttable;

extern mdesc commandtable[];
extern fdesc ifunctiontable[];

int xcommands;
int xfunctions;
int constants;
char * namebuff;
char * parambuff;
char * ifacebuff;
int namelen;
int ifacebufflen;
int SCI_ADDTEXT;
int SCI_APPENDTEXT;
int execerror;

HINSTANCE instance;
CMacroCommunicator * cmc;

extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	if (dwReason==DLL_PROCESS_ATTACH){
		dynmacro=NULL;
		instance=hInstance;
		xcommandtable=NULL;
		xfunctiontable=NULL;
		constanttable=NULL;
		namebuff=NULL;
		parambuff=NULL;
		ifacebuff=NULL;
		debugging=false;
		preparecopydata=false;
	}
	if (dwReason==DLL_PROCESS_DETACH){
		if (xcommandtable!=NULL) delete[] xcommandtable;
		if (xfunctiontable!=NULL) delete[] xfunctiontable;
		if (constanttable!=NULL) delete[] constanttable;
		if (namebuff!=NULL) delete[] namebuff;
		if (parambuff!=NULL) delete[] parambuff;
		if (ifacebuff!=NULL) delete[] ifacebuff;
	}
	return 1;  
}

__declspec( dllexport ) macro * CreateMacro()
{
	return new (macro);
}

__declspec( dllexport ) void DeleteMacro(macro * m)
{
	delete m;
}

__declspec( dllexport ) void PrepareCopyData(bool p)
{
	preparecopydata=p;
}

__declspec( dllexport ) void SetCommunicator(CMacroCommunicator * maccom)
{
	cmc=maccom;
}
__declspec( dllexport ) bool Compile(macro * m,macro * mroot,char * macbeg,char ** macend){
	macros=mroot;
	m->Compile(macbeg,macend);
	return !m->gotcompileerror;
}

__declspec( dllexport ) char * Decompile(macro * m,macro * mroot){
	macros=mroot;
	return m->Decompile();
}

__declspec( dllexport ) char * Dumpcode(macro * m) {
	return m->DumpCode();
}

__declspec( dllexport )void Play(macro * m){
	calling=0;
	stopped=false;
	debugging=false;
	m->Play();
}

__declspec( dllexport )void Debug(macro * m){
	calling=0;
	stopped=false;
	debugging=true;
	m->Play();
	debugging=false;
}
__declspec( dllexport )char * GetValueOf(macro * m,char * varname) {
	return m->GetValue(varname);
}
__declspec( dllexport )char * GetValueOfExpr(macro * m,char * expr) {
	return m->GetExpressionValue(expr);
}

__declspec( dllexport )void Record(macro * m,int message,int wParam,int lParam){
	m->Record(message,wParam,lParam);
}


__declspec( dllexport )char * GetMacroKeyWords(int flag){

	char * buffer = new char[2000+namelen];
	char * bufend=buffer;
	int i;
	if ((flag & 1)==1) {
		//internal stuff
		i=0;
		do{
			bufend+=sprintf(bufend,commandtable[i].desc);
			*bufend++=' ';
			i++;
		}while (commandtable[i-1].message!=103);//MAC_STRING);

		i=0;
		do{
			bufend+=sprintf(bufend,ifunctiontable[i].desc);
			*bufend++=' ';
			i++;
		}while (ifunctiontable[i-1].message!=69);//TOK_SECOND
	}
	if ((flag & 2)==2) {
		//scintilla stuff
		i=0;
		do{
			bufend+=sprintf(bufend,xcommandtable[i].desc);
			*bufend++=' ';
			i++;
		}while(i<=xcommands);

		
		i=0;
		do{
			bufend+=sprintf(bufend,xfunctiontable[i].desc);
			*bufend++=' ';
			i++;
		}while(i<=xfunctions);

		i=0;
		do{
			bufend+=sprintf(bufend,constanttable[i].desc);
			*bufend++=' ';
			i++;
		}while (i<=constants);
	}
	*bufend='\0';

	return buffer;
}

__declspec( dllexport )int GetCommandCode(char * command){
	int i=0;
	do{
		if (!lstrcmp(command,xcommandtable[i].desc))
			return xcommandtable[i].message;
		i++;
	}while (i<=xcommands);
	i=0;
	do{
		if (!lstrcmp(command,xfunctiontable[i].desc))
			return xfunctiontable[i].message;
		i++;
	}while (i<=xfunctions);
	return 0;
}


int typfromletter(char * c){

	if (*c=='I' || *c=='i')
		return TYP_INT;
	else if (*c=='S')
		return TYP_STRING;
	else
		return 0;
}


static char TypeLetter(char * type)
{
	if (*type=='\0' || !strncmp(type,"void",4))
		return '0';
	else if (!strncmp(type,"int",3) ||
			!strncmp(type,"bool",4)||
			!strncmp(type,"position",8)||
			!strncmp(type,"colour",6)||
			!strncmp(type,"keymod",6))
		return 'I';
	else if (!strncmp(type,"stringresult",12))
		return 'R';
	else if (!strncmp(type,"string",6))
		return 'S';
	else 
		return '?';

}

static char GetFeature(char * s){
	if (!strncmp(s,"cat",3))
		return 'c';
	if (!strncmp(s,"fun",3))
		return 'f';
	if (!strncmp(s,"get",3))
		return 'g';
	if (!strncmp(s,"set",3))
		return 's';
	if (!strncmp(s,"val",3))
		return 'v';
	if (!strncmp(s,"evt",3))
		return 'e';
	return '?';
}

__declspec( dllexport )bool LoadIface(char * ifacename){

	int tlen;				//iface file text length
	char * name;			//temp pointer to name
	char * nm;				//namebuffer write pointer
	char * pr;				//parameterbuffer write pointer
	char * dst0;			//temp copy of dst
	char * src;				//source when parsing/copying
	char * dst;				//destination when parsing copying
	char * codeplace;		//placeholder for command code
	int nlen;				//parsed name length
	char t;					//return type
	char t1;				//parameter 1 type				
	char t2;				//parameter 2 type
	char f;					//feature 					
	int codevalue;			//parsed command code value
	bool mrecord;			//true if "MacroRecord" evt
	bool patcht1;			//Eliminate first parm for some calls
	bool goodfield;			//Set to false to eliminate bad line

	xcommands=0;
	xfunctions=0;
	constants=0;
	namelen=0;

	//read iface file
	FILE *fp = fopen(ifacename, "rb");
		if (!fp) return false;
		fseek(fp, 0, SEEK_END);
		tlen = ftell(fp);
		ifacebuff=new char[tlen+1];
		fseek(fp, 0, SEEK_SET);
		fread(ifacebuff, 1, tlen, fp);
	fclose(fp);
	*(ifacebuff+tlen)='\0';
	src=ifacebuff;
	// remove tabs and \r
	while (*src!='\0'){
		if (*src=='\t') *src=' ';
		if (*src=='\r') *src='\n';
		src++;
	}

	// first pass to compact and count commands/functions/constants
	src=ifacebuff;
	dst=ifacebuff;
	goodfield=true; //no field : no error ...

	while (*src!='\0'){
		mrecord=false;
		patcht1=false;
		if (!goodfield){
			dst=dst0;
			while (*src!='\n' && *src!='\0') src++;
		}
		goodfield=false; // will be set to true when parsing is useful and succeeds
		dst0=dst;
		//Find first non space
		while (*src==' ' || *src=='\n') src++;
		if (*src=='\0') break;
		// get feature
		char f=GetFeature(src);
		src+=3;
		// cat, comment, or unsupported
		if (f=='c' || f=='?') continue;
		//store feature
		*dst++=f;
		while (*src==' ') src++;
		// for fun get set : get type
		if (f!='v'){
			t=TypeLetter(src);
			while (isalpha(*src)) src++;
			while (*src==' ' || *src=='\t') src++;
			if (t=='?') continue;
			*dst++=t;
			codeplace=dst;
			dst+=4; //space for code
		}
		
		// for all : get name
		nlen=0;
		char * npos=src;
		while (isalnum(*src) || *src=='_'){
			*dst++=*src++;nlen++;
		}
		if (!strncmp(npos,"MacroRecord",nlen) 
			|| !strncmp(npos,"CharAdded",nlen) 
			|| !strncmp(npos,"Modified",nlen)) 
			mrecord=true; // used notifications
		else if (!strncmp(npos,"AddText",nlen) 
		 || !strncmp(npos,"AppendText",nlen) 
		 || !strncmp(npos,"GetText",nlen) 
		 || !strncmp(npos,"GetCurLine",nlen))
//		 || !strncmp(npos,"WM_GetText",nlen))
			// for these functions/commands, the first parameter
			// is an int containing the length of the second (string) parameter
			// The Macro compiler won't expect the first parameter.
			// When executed, the length will be calculated.
			patcht1=true;


		if (!strncmp(npos,"WM_",3) || !strncmp(npos,"EM_",3))
			continue;
		//only three events are needed 
		if (!mrecord && f=='e') continue;
		// skip spaces
		while (*src==' ' && *src!='\0') src++;
		if (*src!='=') continue;
		// = sign for constants
		if (f=='v') *dst++=*src++;
		// skip spaces
		while (*src==' ' && *src!='\0') src++;
		if (*src=='\0') continue;
		// for constants
		if (f=='v'){
			while (*src!=' ' && *src!='\n' && *src!='\0') *dst++=*src++;
			*dst++='\n';
			namelen+=(nlen+1);
			constants++;
			goodfield=true;
			continue;
		}
		// for fun get set
		src++;
		int cod=atoi(src);
		*((int *)codeplace)=cod;
		while (isdigit(*src)) src++;

		if (*src!='(') continue;

		//copy '('
		*dst++=*src++;
		if (!mrecord){
			while (*src==' ') src++;
			if (*src=='\0') continue;
			if (*src==','){
				t1='0';
				*dst++=t1;
			}else{
				t1=TypeLetter(src);
				if (patcht1) t1='0';
				*dst++=t1;
				//skip type & space
				while(isalpha(*src)) src++;
				while(*src==' ') src++;
				//copy name
				if (patcht1)
					while (isalnum(*src) || *src=='_') src++;	
				else
					while (isalnum(*src) || *src=='_') *dst++=*src++;
				//skip space
				while(*src==' ') src++;
			}
		
			//copy ','
			*dst++=*src++;
			while (*src==' ') src++;
			if (*src==')'){
				t2='0';
				*dst++=t2;
			}else{
				t2=TypeLetter(src);
				*dst++=t2;
				//skip type & space
				while(isalpha(*src)) src++;
				while(*src==' ') src++;
				//copy name
				while (isalnum(*src) || *src=='_') *dst++=*src++;
				//skip space
				while(*src==' ') src++;
			}
		}else{
			t1='0';
			t2='0';
			*dst++=t1;
			*dst++=',';
			*dst++=t2;
		}
		while (*src!=')') src++;

		if (*src!=')') continue;
		*dst++=*src++;
		if (t1=='?' || t2=='?' || (t1=='R' && t2=='R')) continue;
	
		if (t=='0' && t2!='R'){
			xcommands++;
		}else{
			xfunctions++;
		}
		*dst++='\n';
		namelen+=(nlen+1);
		goodfield=true;
	}
		
	*dst='\0';		
	ifacebufflen=dst-ifacebuff;

	parambuff=new char[(xcommands+xfunctions+3)*5];
	xcommandtable =new mdesc[xcommands+1];
	xfunctiontable=new fdesc[xfunctions+1];
	constanttable=new cdesc[constants+3];
	namebuff=new char[namelen+1];


	//Second pass to build tables

	*namebuff='\0';
	nm=namebuff;
	pr=parambuff;
	src=ifacebuff;

	constanttable[0].desc="true";
	constanttable[0].value=1;
	constanttable[1].desc="false";
	constanttable[1].value=0;
	constants=1;
	xcommands=-1;
	xfunctions=-1;
	
	while (*src!='\0'){
		f=*src++;
		if (f=='v'){
			name=src;
			while(*src!='=') src++;			//skip name
			nlen=src-name;

			src++;							//skip '='
			if (isdigit(*src)) 
				sscanf(src,"%i",&codevalue);
			else
				sscanf(src,"%x",&codevalue);
			while (*src!='\n') src++;	//skip code
			src++;						//skip \n
			constants++;
			constanttable[constants].desc=nm;
			constanttable[constants].value=codevalue;
		}else{
			t=*src++;
			codevalue=*((int *)src);
			src+=4;
			name=src;
			while(*src!='(') src++;
			nlen=src-name;
			src++;										//skip '('
			t1=*src++;									//read type1
			while (*src!=',') src++;					//skip name1
			src++;										//skip ','
			t2=*src++;									//read type2
			while (*src!=')') src++;					//skip name2
			src+=2; //skip ')' & \n

			if (t=='0' && t1!='R' && t2!='R'){
				xcommands++;
				xcommandtable[xcommands].desc=nm;
				xcommandtable[xcommands].message=codevalue;
				xcommandtable[xcommands].parms=pr+1;		//0 return value
				*pr++='0';		//no return value
				*pr++=t1;
				*pr++=t2;
				*pr++='i'; //repetition
			//returning int or 1 string in p2
			}else if ((t=='I') || (t2=='R')){
				xfunctions++	;
				xfunctiontable[xfunctions].desc=nm;
				xfunctiontable[xfunctions].message=codevalue;
				xfunctiontable[xfunctions].parms=pr;
				if (t2=='R')
					*pr++='S';
				else
					*pr++='I';
				*pr++=t1;
				*pr++=t2;
			}
			*pr++='\0';			//parameter ending 0
		}
		strncpy(nm,name,nlen);
		nm+=(nlen+1);
		*(nm-1)='\0';
	}
	SCI_ADDTEXT=GetCommandCode("AddText");
	SCI_APPENDTEXT=GetCommandCode("AppendText");
	return true;
}


void Dump1internalCommand(char * desc,char * parms,char ** linebuff,char ** rtype)
{
	char * lptr=*linebuff;
	
	*lptr='\0';
	strcpy(lptr,desc);lptr+=strlen(lptr);
	char * ttyps=parms;
	if (*ttyps=='\0') {
		*rtype="";
		return;
	}
	switch (*ttyps){
		case '0' : *rtype=":void   ";break;
		case 'I' : *rtype=":int    ";break;
		case 'S' : *rtype=":string ";break;
	}	
	ttyps++;
	if (*ttyps=='\0') return;
	*lptr++='(';
	char * lptr0=lptr;
	while (*ttyps!='\0') {
		switch (*ttyps){
			case '0':break;
			case 'I' : strcpy(lptr," int ");lptr+=5;break;
			case 'S' : strcpy(lptr," string ");lptr+=8;break;
			//case 'R' : rettyp='S';break;
		}
		ttyps++;
		if (*(lptr-1)==',') lptr--;
		if (*ttyps=='\0')
			*lptr++=')';
		else if (lptr!=lptr0) 
			*lptr++=',';
	}

	*lptr='\0';
}



char * Dump1Command(char * iiptr,char ** linebuff,char ** rtype)
{
	char * lptr=*linebuff;
	char ptyp;
	char * iptr=iiptr;
	char kind=*iptr++;
	
	*lptr='\0';
	if (kind=='v'){
		while (*iptr!='\n') *lptr++=*iptr++;
		*rtype=":";
	} else {
		char rettyp=*iptr++;
		iptr+=4;

		while (*iptr!='(') *lptr++=*iptr++; 
		*lptr++=*iptr++;
		ptyp=*iptr++;
		switch (ptyp){
			case '0':break;
			case 'I' : strcpy(lptr,"int ");lptr+=4;break;
			case 'S' : strcpy(lptr,"string ");lptr+=7;break;
		}
		while (*iptr!=',') *lptr++=*iptr++; 
		if (ptyp!='0') *lptr++=*iptr;
		iptr++;
		switch (*iptr++){
			case '0':break;
			case 'I' : strcpy(lptr,"int ");lptr+=4;break;
			case 'S' : strcpy(lptr,"string ");lptr+=7;break;
			case 'R' : rettyp='S';break;
		}
		if (*(iptr-1)=='R')
			while (*iptr!=')') *iptr++; 
		else
			while (*iptr!=')') *lptr++=*iptr++; 
		if (*(lptr-1)==',') lptr--;
		*lptr++=*iptr++;

		switch (rettyp){
			case '0' : *rtype=":void   ";break;
			case 'I' : *rtype=":int    ";break;
			case 'S' : *rtype=":string ";break;
		}	
	}
	*lptr='\0';

	return iptr;
}

__declspec( dllexport )char * GetToolTip(char * fname,macro * macs){
	char * linebuff=new char[500];
	char * iptr=ifacebuff;
	char * rtype;
	int i;
	mdesc * m;
	fdesc * f;

	int ltip=strlen(fname);
	// Search
	*linebuff='\0';
	while (iptr<ifacebuff+ifacebufflen){
		if (*iptr=='\0') break;
		if (*iptr!='v') {
			if (!strncmp(iptr+6,fname,ltip)) {
				Dump1Command(iptr,&linebuff,&rtype);
				strcat(linebuff,rtype);
				break;
			}
		} 
		iptr+=6;
		while (*iptr!='\n' && *iptr!='\0') iptr++;
		iptr++;
	}
	if (*linebuff=='\0'){
		i=0;
		do{
			m=&commandtable[i];
			if (!strncmp(m->desc,fname,ltip)) {
				Dump1internalCommand(m->desc,m->parms,&linebuff,&rtype);
				strcat(linebuff,rtype);
				break;
			}
			i++;
		}while (m->message!=2);//MAC_DUMP);	
	}

	if (*linebuff=='\0'){
		i=0;
		do{
			f=&ifunctiontable[i];
			if (!strncmp(f->desc,fname,ltip)) {
				Dump1internalCommand(f->desc,f->parms,&linebuff,&rtype);
				strcat(linebuff,rtype);
				break;
			}
			i++;
		}while (f->message!=69);//TOK_SECOND;		
	}
	
	if (*linebuff=='\0'){
		macro * mac=macs;
		while (mac!=NULL) {
			if (!strncmp(mac->name,fname,ltip)) {
				mac->DecompileTitle(linebuff);
				break;
			}
			mac=mac->next;
		}

	}
	return linebuff;
}


void AddTocbuff(char ** cbuff,char ** linebuff,unsigned int * tlen,int * withnumbers,char * rtype,int * lnum)
{
	if (strlen(*cbuff)+strlen(*linebuff)+25>*tlen){
		*tlen+=1000;
		char * newbuff=new char[*tlen];
		memcpy(newbuff,*cbuff,strlen(*cbuff)+1);
		delete(*cbuff);
		*cbuff=newbuff;
	}
	(*lnum)++;
	if (*withnumbers)
		sprintf(*cbuff+strlen(*cbuff),"%i\t: ",*lnum);
	sprintf(*cbuff+strlen(*cbuff),"%s %s\n",rtype+1,*linebuff);
}


char * DumpCommands(int kind,int withnumbers,macro * mptr)
{
	char * cbuff=new char[2000];
	unsigned int tlen=2000;
	char * linebuff =new char [1000];
	char c;
	char f;
	char * iptr=ifacebuff;
	*cbuff='\0';
	int skip;
	int lnum=0;
	char * rtype;
	int i;
	mdesc * m;
	fdesc * fd;

	if ((kind & 15)!=0) {
		while (iptr<ifacebuff+ifacebufflen){
			f=*iptr;
			c=*(iptr+1);
			if (f=='\0') break;
			skip=0;
			if (f=='e') { 
				skip=6;
			} else if (f=='v') {
				if (!(kind & 1)) skip=1;
			} else {
				if (!(kind & 2) && c=='0') skip=6;
				if (!(kind & 4) && c=='I') skip=6;
				if (!(kind & 8) && c=='S') skip=6;
			}

			if (skip==0){
				iptr=Dump1Command(iptr,&linebuff,&rtype);
				AddTocbuff(&cbuff,&linebuff,&tlen,&withnumbers,rtype,&lnum);
			} else {
				iptr+=skip;
				while (*iptr!='\n' && *iptr!='\0') iptr++;
			}
			iptr++;
		}
	}
	if ((kind & 16)==16){
		i=0;
		do{
			m=&commandtable[i];
			Dump1internalCommand(m->desc,m->parms,&linebuff,&rtype);
			i++;
			AddTocbuff(&cbuff,&linebuff,&tlen,&withnumbers,rtype,&lnum);
		}while (m->message!=2);//MAC_DUMP);	
	}
	if ((kind & 32)==32){
		i=0;
		do{
			fd=&ifunctiontable[i];
			Dump1internalCommand(fd->desc,fd->parms,&linebuff,&rtype);
			i++;
			AddTocbuff(&cbuff,&linebuff,&tlen,&withnumbers,rtype,&lnum);
		}while (fd->message!=69);//TOK_SECOND);	
	}
	if ((kind & 64)==64){
		macro * mac=mptr;
		while (mac->prev!=NULL) {
			mac=mac->prev;
		}
		while (mac!=NULL) {
			mac->DecompileTitle(linebuff);
			AddTocbuff(&cbuff,&linebuff,&tlen,&withnumbers,":",&lnum);
			mac=mac->next;
		}
	}
	delete linebuff;
	return cbuff;

}
